import tensorflow as tf
from tensorflow_probability import distributions as tfd
import numpy as np
from careless.models.priors.base import Prior
from careless.models.merging.surrogate_posteriors import RiceWoolfson
import math


class ReferencePrior():
    """
    A Prior class with a `log_prob` implementation that returns zeros for unobserved miller indices.
     - This class is not meant to be used directly. 
     - Extensions of this class must set the `base_dist` attribute with a tfd.Distribution or similar object with
       a `.log_prob(values)` method.
    """
    base_dist = None
    def __init__(self, observed=None):
        if observed is None:
            self.idx = None
        else:
            idx = tf.where(observed)
            self.idx = tf.reshape(idx, (-1,))

    def mean(self):
        """ This just passes through to self.base_dist. """
        return self.base_dist.mean()

    def stddev(self):
        """ This just passes through to self.base_dist. """
        return self.base_dist.stddev()

    def log_prob(self, values):
        if self.idx is None:
            return self.base_dist.log_prob(values)
        obs = tf.gather(values, self.idx, axis=-1)
        log_prob = self.base_dist.log_prob(obs)

        return tf.transpose(tf.scatter_nd(
            self.idx[...,None], 
            tf.transpose(log_prob), 
            values.shape[::-1]
        ))

class LaplaceReferencePrior(ReferencePrior):
    """
    A Laplacian prior distribution centered at empirical structure factor amplitudes derived from a conventional experiment.
    """
    def __init__(self, Fobs, SigFobs, observed=None):
        """
        Parameters
        ----------
        Fobs : array
            numpy array or tf.Tensor containing observed structure factors amplitudes from a reference structure.
        SigFobs : array
            numpy array or tf.Tensor containing error estimates for structure factors amplitudes from a reference structure.
        observed : array (optional)
            boolean numpy array or tf.Tensor which has True for all observed miller indices.
        """
        super().__init__(observed)
        loc = np.array(Fobs, dtype=np.float32)
        scale = np.array(SigFobs, dtype=np.float32)/math.sqrt(2.)
        self.base_dist = tfd.Laplace(loc, scale)

class NormalReferencePrior(ReferencePrior):
    """
    A Normal prior distribution centered at empirical structure factor amplitudes derived from a conventional experiment.
    """
    def __init__(self, Fobs, SigFobs, observed=None):
        """
        Parameters
        ----------
        Fobs : array
            numpy array or tf.Tensor containing observed structure factors amplitudes from a reference structure.
        SigFobs : array
            numpy array or tf.Tensor containing error estimates for structure factors amplitudes from a reference structure.
        observed : array (optional)
            boolean numpy array or tf.Tensor which has True for all observed miller indices.
        """
        super().__init__(observed)
        loc = np.array(Fobs, dtype=np.float32)
        scale = np.array(SigFobs, dtype=np.float32)
        self.base_dist = tfd.Normal(loc, scale)

class StudentTReferencePrior(ReferencePrior):
    """
    A Student's T prior distribution centered at empirical structure factor amplitudes derived from a conventional experiment.
    """
    def __init__(self, Fobs, SigFobs, dof, observed=None):
        """
        Parameters
        ----------
        Fobs : array
            numpy array or tf.Tensor containing observed structure factors amplitudes from a reference structure.
        SigFobs : array
            numpy array or tf.Tensor containing error estimates for structure factors amplitudes from a reference structure.
        dof : float
            degrees of freedom for the student's t distribution.
        observed : array (optional)
            boolean numpy array or tf.Tensor which has True for all observed miller indices.
        """
        super().__init__(observed)
        loc = np.array(Fobs, dtype=np.float32)
        scale = np.array(SigFobs, dtype=np.float32)
        self.base_dist = tfd.StudentT(dof, loc, scale)

class RiceWoolfsonReferencePrior(ReferencePrior):
    """
    A Rice Woolfson prior distribution centered at empirical structure factor amplitudes derived from a conventional experiment.
    """
    def __init__(self, Fobs, SigFobs, centric, observed=None):
        """
        Parameters
        ----------
        Fobs : array
            numpy array or tf.Tensor containing observed structure factors amplitudes from a reference structure.
        SigFobs : array
            numpy array or tf.Tensor containing error estimates for structure factors amplitudes from a reference structure.
        centric : array
            boolean numpy array or tf.Tensor which has True for all centric reflections.
        observed : array (optional)
            boolean numpy array or tf.Tensor which has True for all observed miller indices.
        """
        super().__init__(observed)
        loc = np.array(Fobs, dtype=np.float32)
        scale = np.array(SigFobs, dtype=np.float32)
        self.base_dist = RiceWoolfson(loc, scale, centric)





name = "Likelihood Options"
description = None

args_and_kwargs = (
    (("--studentt-likelihood-dof",), { 
        "help":"Degrees of freedom for student t likelihood function.",
        "type":float, 
        "metavar":'DOF', 
        "default":None,
    }),

    (("--refine-uncertainties",), { 
        "help":"Use Evans' 2011 error model from SCALA to correct uncertainties.",
        "action":'store_true', 
        "default":False,
    }),
)




"""
Filter reflections by image based on CCpred.
"""
import re
import argparse
import numpy as np
import reciprocalspaceship as rs
import gemmi

import matplotlib.pyplot as plt
import seaborn as sns

from careless.io.formatter import get_first_key_of_dtype

class ArgumentParser(argparse.ArgumentParser):
    def __init__(self):
        super().__init__(
            description=__doc__
        )

        # Required arguments
        self.add_argument(
            "mtzs",
            nargs="+",
            help="A mix of *_predictions_#.mtz files and unmerged input files from careless. "
                 "Note: the filenames will be used to interpret what they are. Do not "
                 "use renamed files. The unmerged files must be supplied in the same order"
                 "they were presented to careless.",
        )

        # Optional arguments
        self.add_argument(
            "-m",
            "--method",
            default="weighted",
            choices=["weighted", "pearson", "spearman"],
            help="Method for computing correlation coefficient (spearman or pearson). "
            "Weighted is the default.",
        )

        self.add_argument(
            "-c",
            "--cc-cutoff",
            required=True,
            type=float,
            help="The correlation cutoff. A number between zero and 1."
        )

        self.add_argument(
            "-o",
            default=None,
            help="Output filename base.",
        )

def weighted_pearson_ccfunc(df, iobs='Iobs', ipred='Ipred', sigiobs='SigIobs'):
    x = df[iobs].to_numpy('float32')
    y = df[ipred].to_numpy('float32')
    w = np.reciprocal(np.square(df[sigiobs])).to_numpy('float32')
    return rs.utils.weighted_pearsonr(x, y, w)

def pearson_ccfunc(df, iobs='Iobs', ipred='Ipred'):
    return df[[iobs, ipred]].corr(method='pearson')[iobs][ipred]

def spearman_ccfunc(df, iobs='Iobs', ipred='Ipred'):
    return df[[iobs, ipred]].corr(method='spearman')[iobs][ipred]

def is_predictions_filename(filename : str) -> bool:
    rematch = re.match(r".+predictions_[0-9]+\.mtz$", filename)
    retval = rematch is not None
    return retval

def predictions_id(filename : str) -> int:
    return int(filename[:-4].split('_')[-1])

def run_analysis(args):
    try:
        data_mtzs = [f for f in args.mtzs if not is_predictions_filename(f)]
        predictions_mtzs = [f for f in args.mtzs if is_predictions_filename(f)]
        predictions_mtzs = sorted(predictions_mtzs, key=predictions_id)
    except:
        raise ValueError(
            "Unable to interpret provided filenames {args.mtzs}. Please ensure "
            "predictions files have not been renamed. Unmerged mtzs must be in the"
            " order provided during the original careless run."
        )

    ds = []
    for m in predictions_mtzs:
        _ds = rs.read_mtz(m)
        #non-isomorphism could lead to different resolution for each mtz
        #need to calculate dHKL before concatenating 
        _ds.compute_dHKL(inplace=True)
        _ds['file'] = m
        _ds['Spacegroup'] = _ds.spacegroup.xhm()
        ds.append(_ds)
    ds = rs.concat(ds, check_isomorphous=False)

    grouper = ds.groupby(["file", "image_id"])

    if args.method.lower() == "spearman":
        ccfunc = spearman_ccfunc
    elif args.method.lower() == "weighted":
        ccfunc = weighted_pearson_ccfunc
    elif args.method.lower() == "pearson":
        ccfunc = pearson_ccfunc
    else:
        raise ValueError(f"Unrecognized CC --method, {args.method}")

    result = grouper.apply(ccfunc)
    result = rs.DataSet({"CCpred" : result}).reset_index()
    result['file_id'] = grouper.first()['file_id'].to_numpy()
    result['asu_id'] = grouper.first()['asu_id'].to_numpy()
    result = result[['file', 'file_id', 'asu_id', 'image_id', 'CCpred']]

    for i,m in enumerate(data_mtzs):
        ds = rs.read_mtz(m)
        if args.o is None:
            out = m[:-4] + '_filtered.mtz'
        else:
            out = args.o + f'_{i}.mtz'

        batch_key = get_first_key_of_dtype(ds, 'B')
        image_id = ds.groupby(batch_key).ngroup().to_numpy()
        cc = result[result.file_id == i].iloc[image_id].CCpred.to_numpy()
        idx = (cc >= args.cc_cutoff)
        ds[idx].copy().write_mtz(out)

def main():
    parser = ArgumentParser().parse_args()
    run_analysis(parser)




name = "Data Interpretation"
description = None



args_and_kwargs = (
    (("--spacegroups", ),  {
        "help":"The spacegroup(s) to use for merging. You may either supply a single spacegroup "
               "which will be used for every input reflection file or a comma-separated list of "
               "spacegroups with the same length as the number of reflection files. For example "
               '--spacegroups="P 21 21 21" or --spacegroups="P 21 21 21,P 1 21 1"',
        "type":str, 
        "default" : None,
    }),

    (("--image-key", ),  {
        "help":"The name of the key indicating image number for each data set. "
               "If no key is given, careless will use the first key with the BATCH dtype.",
        "type":str, 
        "default" : None,
    }),

    (("--intensity-key", ),  {
        "help":"What key to use for reflection intensities. "
               "If no key is given, careless will use the first key with the intensity dtype.",
        "type":str, 
        "default" : None,
    }),

    (("--uncertainty-key", ),  {
        "help":"What key to use for reflection error estimates. "
               "If no key is given, careless will first check for a key beginning with 'SIG' or 'Sig' "
               "which matches the intensity key (e.g. Iobs -> SigIobs). "
               "Failing that, careless will ust first key with the StdDev dtype.",
        "type":str, 
        "default" : None,
    }),

    (("--anomalous",), { 
        "help":f"If this flag is supplied, Friedel mates will be kept separate.", 
        "action":'store_true', 
        "default":False,
    }),

    (("--separate-files",), {
        "help":"Use this flag to produce a separate output for each input mtz." 
               "In this mode, the data will be 'scaled' together and 'merged' separately." 
               "The default is to merge all the files into a single output.", 
        "action" : "store_true",
        "default": False,
    }),
)



import tensorflow as tf

def sanitize_tensor(tensor, replacement_val=0.):
    """Replace infinite entries `replacement_val`."""
    return tf.where(~tf.math.is_finite(tensor), replacement_val*tf.ones_like(tensor), tensor)




import argparse
import numpy as np 
from os.path import exists

class EnvironmentSettingsMixin(argparse.ArgumentParser):
    """
    This will automagically set tensorflow environment variables when parse_args is called.
    """
    def parse_args(self, *args, **kwargs):
        parser = super().parse_args(*args, **kwargs)

        from os import environ
        environ["TF_USE_LEGACY_KERAS"] = "1" #use keras 2
        if parser.tf_debug:
            # This is very noisy
            environ['TF_CPP_MIN_LOG_LEVEL'] = "1"
        else:
            # This is very quiet
            environ['TF_CPP_MIN_LOG_LEVEL'] = "3"

        import tensorflow as tf
        np.random.seed(parser.seed)
        tf.random.set_seed(parser.seed)

        #Disable the GPU if requested. This can be useful for training multiple models at the same time
        if parser.disable_gpu:
            tf.config.set_visible_devices([], 'GPU')
        #For multi-gpu machines allocate only the zeroth GPU.
        else:
            #Set active GPU
            gpus = tf.config.experimental.list_physical_devices('GPU')
            gpu_id = parser.gpu_id

            try:
                gpu = gpus.pop(gpu_id)
                if not parser.disable_memory_growth:
                    tf.config.experimental.set_memory_growth(gpu, True)
                tf.config.experimental.set_visible_devices(gpu, 'GPU')
            except:
                tf.config.experimental.set_visible_devices([], 'GPU')

        return parser

class CustomParser(EnvironmentSettingsMixin):
    """
    A custom ArgumentParser with parse_args overloaded in order to 
     - Set tensorflow environment variables
     - Detect conflicting arguments and raise an informative error
    """
    def _validate_input_files(self, parser):
        if parser.type == 'devices':
            return
        for inFN in parser.reflection_files:
            if not exists(inFN):
                self.error(f"Unmerged reflection file {inFN} does not exist")
            elif inFN.endswith(".mtz") or inFN.endswith(".stream"):
                continue
            self.error(
                f"Could not determine filetype for reflection file, {inFN}." 
                 "Please make sure your files end in '.mtz' or '.stream' as"
                 " appropriate."
                )

    def parse_args(self, *args, **kwargs):
        parser = super().parse_args(*args, **kwargs)
        self._validate_input_files(parser)
        return parser

import re
import textwrap
class CustomFormatter(argparse.HelpFormatter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._whitespace_matcher=re.compile("\n(?!\n)")

    def _fill_text(self, text, width, indent):
        #First replace single newlines with a space
        text = re.sub(r'(?!>\n)\n(?!\n)', '', text)
        return textwrap.fill(text, width, initial_indent=indent, subsequent_indent=indent, replace_whitespace=False, drop_whitespace=False)

description = """
Scale and merge crystallographic data by \n\n\n approximate inference.
"""
parser = CustomParser(description=description, formatter_class=CustomFormatter)

# Add --version argument
import careless
parser.add_argument("--version", action="version", version=f"careless {careless.__version__}")

subs = parser.add_subparsers(title="Experiment Type", required=True, dest="type")
mono_sub = subs.add_parser("mono", help="Process monochromatic diffraction data.", formatter_class=CustomFormatter)
poly_sub = subs.add_parser("poly", help="Process polychromatic, 'Laue', diffraction data.", formatter_class=CustomFormatter)
devices_sub = subs.add_parser("devices", help="Print available physical devices", formatter_class=CustomFormatter)

from careless.args import required,poly,groups

for args,kwargs in required.args_and_kwargs:
    mono_sub.add_argument(*args, **kwargs)
    poly_sub.add_argument(*args, **kwargs)

for args,kwargs in poly.args_and_kwargs:
    poly_sub.add_argument(*args, **kwargs)

for group in groups:
    if group.name is not None and group.description is not None:
        mono_group = mono_sub.add_argument_group(group.name, group.description)
        poly_group = poly_sub.add_argument_group(group.name, group.description)
    elif group.name is not None:
        mono_group = mono_sub.add_argument_group(group.name)
        poly_group = poly_sub.add_argument_group(group.name)
    else:
        mono_group = mono_sub
        poly_group = poly_sub
    for args,kwargs in group.args_and_kwargs:
        mono_group.add_argument(*args, **kwargs)
        poly_group.add_argument(*args, **kwargs)

# Test needs environment settings options
from careless.args import tf_options
for args,kwargs in tf_options.args_and_kwargs:
    devices_sub.add_argument(*args, **kwargs)



name = "Laue"
description = None

args_and_kwargs = (
    (("-l", "--wavelength-range"), {
        "help":"Minimum and maximum wavelength for harmonic deconvolution in Ångstroms. "
               "If this is not supplied, Harmonics will be predicted out to the minimum and "
               "maximum wavelengths recorded in the mtz.", 
        "type": float, 
        "default": None,
        "nargs":2, 
        "metavar": ('lambda_min', 'lambda_max'),
    }),

    (("-w", "--wavelength-key"),  {
        "help":f"Mtz column name corresponding to the reflections' peak wavelength.", 
        "type":str, 
        "default":'Wavelength',
    }),
)



import numpy as np

def positional_encoding(X, L):
    """
    The positional encoding as defined in the NeRF paper https://arxiv.org/pdf/2003.08934.pdf
      gamma(p) = (sin(2**0*pi*p), cos(2**0*pi*p), ..., sin(2**(L-1)*pi*p), cos(2**(L-1)*pi*p))
    Wherein p represents an arbitrary batched set of vectors computed by normalizing X between
    between -1 and 1
    """
    p = 2.*(X - X.min(-2)) / (X.max(-2) - X.min(-2)) - 1.
    L = np.arange(L, dtype=X.dtype)
    f = np.pi*2**L
    fp = (f[...,None,:]*p[...,:,None]).reshape(p.shape[:-1] + (-1,))
    return np.concatenate((
        np.cos(fp),
        np.sin(fp),
    ), axis=-1)





"""
Plot training history from careless output.
"""

import seaborn as sns
import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from argparse import ArgumentParser,RawTextHelpFormatter


class ArgumentParser(ArgumentParser):
    def __init__(self, *args, **kwargs):
        super().__init__(
            formatter_class=RawTextHelpFormatter, 
            description=__doc__
        )
        self.add_argument('history_csv', type=str, help='A _history.csv file generated by careless')
        self.add_argument('-o', type=str, default=None, help='Output image file name')
        self.add_argument('-s', '--show', action='store_true')


def plot_history(history_csv, step_key='step'):
    df = pd.read_csv(history_csv).melt(step_key)
    val = df.variable.str.endswith('_val')
    df['variable'] = df.variable.str.removesuffix('_val')
    df['Set'] = np.array(['Train', 'Test'])[val.to_numpy('int')]
    sns.lineplot(
        data=df,
        x='step',
        y='value',
        hue='variable',
        style='Set',
        palette='Dark2',
    )
    plt.semilogy()
    plt.grid(which='both', axis='both', linestyle='--')


def run_analysis(parser):
    plot_history(parser.history_csv)

    if parser.o is not None: 
        plt.savefig(parser.o, dpi=300)
    if parser.show:
        plt.show()

def main():
    parser = ArgumentParser()
    parser = parser.parse_args()
    run_analysis(parser)




from careless.models.base import BaseModel
from careless.utils.shame import sanitize_tensor
from careless.models.merging.surrogate_posteriors import TruncatedNormal
import tf_keras as tfk
from tqdm.autonotebook import tqdm
import tensorflow_probability as tfp
import tensorflow as tf
import numpy as np


class VariationalMergingModel(tfk.models.Model, BaseModel):
    """
    Merge data with a posterior parameterized by a surrogate distribution.
    """
    def __init__(self, surrogate_posterior, prior, likelihood, scaling_model, mc_sample_size=1, kl_weight=None, scale_kl_weight=None, scale_prior=None):
        """"
        Parameters
        ----------
        surrogate_posterior : tfd.Distribution
            A surrogate posterior distribution to use. 
            If non is supplied, the default truncated normal distribution will be used. 
            Any posteriors passed in with this arg must have all properly transformed parameters in their
            `self.trainable_variables` iterable.  Use `tfp.util.TransformedVariable` to ensure positivity constraints
            where applicable.
        prior : distribution
            Prior distribution on merged, normalized structure factor amplitudes. 
            Either a Distribution from tensorflow_probability.distributions or 
            a Prior from careless.models.distributions. This must implement .log_prob. 
            This distribution must have an `event_shape` equal to `np.max(miller_ids) + 1`.
        likelihood : careless.models.likelihood.Likelihood
            This is a Likelihood object from careless.
        scaling_model : careless.models.base.BaseModel
            An instance of a class from carless.model.scaling 
        mc_sample_size : int (optional)
            This sets how many reparameterized samples will be used to compute the loss function.
        """
        super().__init__()
        self.prior = prior
        self.surrogate_posterior = surrogate_posterior
        self.likelihood = likelihood
        self.scaling_model = scaling_model
        self.mc_sample_size = mc_sample_size
        self.kl_weight = kl_weight
        self.scale_kl_weight = scale_kl_weight
        self.scale_prior = scale_prior

    def scale_mean_stddev(self, inputs):
        """
        Compute the moments of the posterior of reflection observation scale factors. 

        Parameters
        ----------
        inputs : data
            inputs is a data structure like [refl_id, image_id, metadata, intensity, uncertainty]. 
            This can be a tf.DataSet, or a group of tensors. 

        Returns
        -------
        mean : np.array
            A numpy array containing the mean value of the scale predicted by the model for each input. 
        stddev : np.array
            A numpy array containing the standard deviation of the scale predicted by the model for each input. 
            This is a reasonable estimate of the uncertainty of the model about each input.
        """
        refl_id = self.get_refl_id(inputs)

        scale_dist = self.scaling_model(inputs)
        mean = scale_dist.mean().numpy()
        stddev = scale_dist.stddev().numpy()

        # We need to convolve the predictions if this is laue data
        from careless.models.likelihoods.laue import LaueBase
        if isinstance(self.likelihood, LaueBase):
            likelihood = self.likelihood(inputs)
            mean = likelihood.convolve(mean)
            stddev = np.sqrt(likelihood.convolve(stddev * stddev))

        return mean, stddev

    def prediction_mean_stddev(self, inputs):
        """
        Parameters
        ----------
        inputs : data
            inputs is a data structure like [refl_id, image_id, metadata, intensity, uncertainty]. 
            This can be a tf.DataSet, or a group of tensors. 

        Returns
        -------
        mean : np.array
            A numpy array containing the mean value predicted by the model for each input. 
        stddev : np.array
            A numpy array containing the standard deviation predicted by the model for each input. 
            This is a reasonable estimate of the uncertainty of the model about each input.
        """
        refl_id = self.get_refl_id(inputs)
        #Let's actually return the expected value of the data under the current model
        #This is <F**2.>
        scale_dist = self.scaling_model(inputs)
        f2 = tf.square(self.surrogate_posterior.mean()) + tf.square(self.surrogate_posterior.stddev())
        iexp = scale_dist.mean() * tf.gather(f2, tf.squeeze(refl_id, axis=-1), axis=-1)
        iexp = iexp.numpy()

        from scipy.stats import truncnorm
        q = self.surrogate_posterior
        f4 = q.moment_4(method='scipy')

        s2 = np.square(scale_dist.mean().numpy()) + np.square(scale_dist.stddev().numpy())
        # var(I) = <I^2> - <I>^2
        # <I^2> = <F^4><Sigma^2>
        ivar = f4[np.squeeze(refl_id)]*s2 - iexp*iexp

        # We need to convolve the predictions if this is laue data
        from careless.models.likelihoods.laue import LaueBase
        if isinstance(self.likelihood, LaueBase):
            likelihood = self.likelihood(inputs)
            iexp = likelihood.convolve(iexp)
            ivar = likelihood.convolve(ivar)
            iexp,ivar = iexp.numpy(),ivar.numpy()

        return iexp,np.sqrt(ivar)

    def add_kl_div(self, posterior, prior, samples=None, weight=1., reduction='sum', name="KLDiv"):
        try:
            kl_div = posterior.kl_divergence(prior)
        except:
            NotImplementedError
            kl_div = posterior.log_prob(samples) - prior.log_prob(samples)

        if reduction == 'sum':
            kl_div = tf.reduce_sum(kl_div) / self.mc_sample_size
        elif reduction == 'mean':
            kl_div = tf.reduce_mean(kl_div) 
        else:
            kl_div = reduction(kl_div)

        self.add_loss(weight * kl_div)
        self.add_metric(kl_div, name=name)
        return kl_div

    def call(self, inputs):
        """
        Parameters
        ----------
        inputs : data
            inputs is a data structure like [refl_id, image_id, metadata, intensity, uncertainty]. 
            This can be a tf.DataSet, or a group of tensors. 

        Returns
        -------
        predictions : tf.Tensor
            Values predicted by the model for this sample. 
        """
        z_f = self.surrogate_posterior.sample(self.mc_sample_size)

        scale_dist = self.scaling_model(inputs)
        z_scale = scale_dist.sample(self.mc_sample_size)

        if self.scale_prior is not None:
            if self.scale_kl_weight is None:
                self.add_kl_div(scale_dist, self.scale_prior, z_scale, weight=self.scale_kl_weight, reduction='sum', name="Σ KLDiv")
            else:
                self.add_kl_div(scale_dist, self.scale_prior, z_scale, weight=1., reduction='mean', name="Σ KLDiv")

        refl_id = self.get_refl_id(inputs)

        ipred = z_scale * tf.square(tf.gather(z_f, tf.squeeze(refl_id, axis=-1), axis=-1))

        likelihood = self.likelihood(inputs)

        ll = likelihood.log_prob(ipred)
        if self.kl_weight is None:
            self.add_kl_div(self.surrogate_posterior, self.prior, z_f, name='F KLDiv', reduction='sum')
            ll = tf.reduce_sum(ll) / self.mc_sample_size
        else:
            self.add_kl_div(self.surrogate_posterior, self.prior, z_f, weight=self.kl_weight, name='F KLDiv', reduction='mean')
            ll = tf.reduce_mean(ll) 

        #Do some keras-y stuff
        self.add_loss(-ll)
        self.add_metric(-ll, name="NLL")

        return ipred

    def train_step_with_gradient_norm(self, data=None):
        """
        Conduct a training step with `data`. This method is the same as tfk.Model.train_step except that it
        tracks the norm of the gradients as well. 
        """
        if data is None:
            x = self.data
        else:
            x = data[0]
        y = self.get_intensities(x)

        # Run forward pass.
        with tf.GradientTape() as tape:
            y_pred = self(x, training=True)
            loss = self.compiled_loss(y, y_pred, regularization_losses=self.losses)

        # Run backwards pass.
        grads = tape.gradient(loss, self.trainable_variables)

        # Compute the L2 norm of the gradients
        grad_norm = tf.linalg.global_norm(grads)

        # Only apply gradients if they are valid
        grads = [tf.where(tf.math.is_finite(g), g, 0.) for g in grads]
        self.optimizer.apply_gradients(zip(grads, self.trainable_variables))

        self.compiled_metrics.update_state(y, y_pred)

        # Collect metrics to return
        return_metrics = {
            "Grad Norm" : grad_norm,
        }
        for metric in self.metrics:
            result = metric.result()
            if isinstance(result, dict):
                return_metrics.update(result)
            else:
                return_metrics[metric.name] = result

        return return_metrics

    def train_model(self, data, steps, message=None, format_string="{:0.2e}", validation_data=None, validation_frequency=10, progress=True, use_custom_train_step=True, jit_compile=None, reduce_retracing=False):
        """
        Alternative to the keras backed VariationalMergingModel.fit method. This method is much faster at the moment but less flexible.
        """
        if use_custom_train_step:
            def train_step(model_and_data):
                model, data = model_and_data
                model.reset_metrics()
                history = model.train_step_with_gradient_norm((data,))
                return history
        else:
            def train_step(model_and_data):
                model, data = model_and_data
                model.reset_metrics()
                history = model.train_step((data,))
                return history

        if not self._run_eagerly:
            train_step = tf.function(
                train_step, reduce_retracing=reduce_retracing, jit_compile=jit_compile
            )

        if validation_data is not None:
            val_scale = len(data[0]) / len(validation_data[0])

        history = {}
        from tqdm import trange
        disable_progress = not progress
        bar = trange(steps, desc=message, disable=disable_progress)
        for i in bar:
            _history = train_step((self, data))
            if validation_data is not None:
                if i%validation_frequency==0:
                    validation_metrics = self.test_on_batch(validation_data, return_dict=True)
                _history['NLL_val'] = val_scale * validation_metrics['NLL']

            pf = {}
            for k,v in _history.items():
                v = float(v)
                pf[k] = format_string.format(v)
                if k not in history:
                    history[k] = []
                history[k].append(v)

            bar.set_postfix(pf)
            if use_custom_train_step:
                if not tf.math.is_finite(_history['Grad Norm']):
                    print("Encountered numerical issues, terminating optimization early!")
                    break
        return history




import tensorflow as tf
import tf_keras as tfk
import numpy as np



class BaseModel(tfk.layers.Layer):
    """ 
    Base class for all models in `careless`. 
    It encodes accessors for the standard format inputs to the model. 
    When extending this class, use the get_{
        metadata,
        refl_id,
        image_id,
        intensities,
        uncertainties,
        wavelength,
        harmonic_id
    } static methods to parse your inputs. This ensures it is easy to
    change the data format in the future.
    """
    input_index = {
        'refl_id'       : 0,
        'image_id'      : 1,
        'file_id'       : 2,
        'metadata'      : 3,
        'intensities'   : 4,
        'uncertainties' : 5,
        'wavelength'    : 6,
        'harmonic_id'   : 7,
    }

    def call(self, inputs):
        raise NotImplementedError(
            "All Scaler classes must implement a call method which accepts inputs "
            "defined by this class.                                               "
        )

    @staticmethod
    def is_laue(inputs : tuple) -> bool:
        """
        Test if the inputs are from Laue or Mono data
        """
        laue_size = BaseModel.get_index_by_name("harmonic_id") + 1
        if len(inputs) >= laue_size:
            return True
        return False

    @staticmethod
    def get_name_by_index(index : int) -> str:
        for k,v in BaseModel.input_index.items():
            if v == index:
                return k
        raise ValueError(
            f"index, {index}, not a valid index. Valid indices are {BaseModel.input_index.values()}."
        )

    @staticmethod
    def get_index_by_name(name):
        if name not in BaseModel.input_index:
            raise ValueError(
                f"name, {name}, not a valid key. Valid keys are {BaseModel.input_index.keys()}."
            )
        return BaseModel.input_index[name]

    @staticmethod
    def get_input_by_name(inputs, name):
        if name not in BaseModel.input_index:
            raise ValueError(
                f"name, {name}, not a valid key. Valid keys are {BaseModel.input_index.keys()}."
            )
        idx = BaseModel.input_index[name]
        try:
            datum = inputs[idx]
        except:
            raise ValueError(
                f"Attempting to gather {name} data from input tensors, {inputs}, with length {len(inputs)} failed."
            )
        if datum.shape[0] == 1:
            datum = tf.squeeze(datum, axis=0)
        return datum

    @staticmethod
    def get_refl_id(inputs):
        """ Given a collection of inputs extract just the reflection_id """
        return BaseModel.get_input_by_name(inputs, 'refl_id')

    @staticmethod
    def get_file_id(inputs):
        """ Given a collection of inputs extract just the file_id """
        return BaseModel.get_input_by_name(inputs, 'file_id')

    @staticmethod
    def get_image_id(inputs):
        """ Given a collection of inputs extract just the image_id """
        return BaseModel.get_input_by_name(inputs, 'image_id')

    @staticmethod
    def get_metadata(inputs):
        """ Given a collection of inputs extract just the metadata """
        return BaseModel.get_input_by_name(inputs, 'metadata')

    @staticmethod
    def get_intensities(inputs):
        """ Given a collection of inputs extract just the intensities """
        return BaseModel.get_input_by_name(inputs, 'intensities')

    @staticmethod
    def get_uncertainties(inputs):
        """ Given a collection of inputs extract just the uncertainty estimates """
        return BaseModel.get_input_by_name(inputs, 'uncertainties')

    @staticmethod
    def get_wavelength(inputs):
        """ Given a collection of inputs extract just the wavelength. This method only applies to Laue data."""
        return BaseModel.get_input_by_name(inputs, 'wavelength')

    @staticmethod
    def get_harmonic_id(inputs):
        """ Given a collection of inputs extract just the harmonic_id. This method only applies to Laue data."""
        return BaseModel.get_input_by_name(inputs, 'harmonic_id')




from careless.models.base import BaseModel


class Likelihood(BaseModel):
    def call(inputs):
        raise NotImplementedError(
            "Likelihoods must implement a call method that returns a `tfp.distribution.Distribution` "
            "or a similar object with a `log_prob` method."
        )




import numpy as np
import tensorflow as tf
import tf_keras as tfk
import reciprocalspaceship as rs
import gemmi
from .asu import ReciprocalASU,ReciprocalASUCollection
from careless.models.base import BaseModel
from careless.models.priors.wilson import WilsonPrior,DoubleWilsonPrior

class DataManager():
    """
    This class comprises various data manipulation methods as well as methods to aid in model construction.
    """
    parser = None
    def __init__(self, inputs, asu_collection, parser=None):
        """
        Parameters
        ----------
        inputs : tuple
        asu_collection : ReciprocalASUCollection
        parser : Namespace (optional)
            A Namespace instance created by careless.parser.parser.parse_args()
        """
        self.inputs = inputs
        self.asu_collection = asu_collection
        self.parser = parser

    @classmethod
    def from_pickle(cls, filename):
        import pickle
        with open(filename, 'rb') as f:
            dm = pickle.load(f)
        return dm

    @classmethod
    def from_mtz_files(cls, filenames, formatter):
        return cls.from_datasets((rs.read_mtz(i) for i in filenames), formatter)

    @classmethod
    def from_stream_files(cls, filenames, formatter):
        return cls.from_datasets((rs.read_crystfel(i) for i in filenames), formatter)

    @staticmethod
    def wilson_sigma(b, dHKL):
        sigma = np.exp(-0.25 * b * np.reciprocal(dHKL*dHKL))
        return sigma

    def get_wilson_sigma(self, b=None):
        if b is None:
            return 1.
        sigma = self.wilson_sigma(b, self.asu_collection.dHKL)
        return sigma

    def get_wilson_prior(self, b=None, k=1.):
        """ Construct a wilson prior with an optional temperature factor, b, appropriate for self.asu_collection. """
        if b is None:
            sigma = 1.
        elif isinstance(b, float):
            sigma = self.get_wilson_sigma(b)
        else:
            raise ValueError(f"parameter b has type{type(b)} but float was expected")
        sigma = sigma * k

        return WilsonPrior(
            self.asu_collection.centric,
            self.asu_collection.multiplicity,
            sigma,
        )

    def get_tf_dataset(self, inputs=None):
        """
        Pack a dataset in the way that keras and careless expect.

        Parameters
        ----------
        inputs : tuple (optional)
            If None, self.inputs will be used
        """
        if inputs is None:
            inputs = self.inputs

        inputs = tuple(inputs)
        iobs = BaseModel.get_intensities(inputs)
        sigiobs = BaseModel.get_uncertainties(inputs)
        packed  = (inputs, iobs, sigiobs)
        tfds = tf.data.Dataset.from_tensor_slices(packed)
        return tfds.batch(len(iobs))

    def get_predictions(self, model, inputs=None, test_value=0):
        """ 
        Extract results from a surrogate_posterior.

        Parameters
        ----------
        model : VariationalMergingModel
            A merging model from careless
        inputs : tuple (optional)
            Inputs for which to make the predictions if None, self.inputs is used.
        test_value : int (optional)
            Optionally change the value of the `test` column used for crossvalidation.
            The default is 0. 

        Returns
        -------
        predictions : tuple
            A tuple of rs.DataSet objects containing the predictions for each 
            ReciprocalASU contained in self.asu_collection
        """
        laue = BaseModel.is_laue(inputs)

        if inputs is None:
            inputs = self.inputs

        refl_id = BaseModel.get_refl_id(inputs)
        asu_id,H = self.asu_collection.to_asu_id_and_miller_index(refl_id)
        asu_id = asu_id.flatten()
        file_id = model.get_file_id(inputs).flatten()
        image_id = model.get_image_id(inputs).flatten()
        if laue:
            harmonic_id = BaseModel.get_harmonic_id(inputs).flatten()
        else:
            harmonic_id = np.arange(len(refl_id))
        h,k,l = H.T

        output = rs.DataSet({
            'H' : rs.DataSeries(h, dtype='H'),
            'K' : rs.DataSeries(k, dtype='H'),
            'L' : rs.DataSeries(l, dtype='H'),
            'harmonic_id' : rs.DataSeries(harmonic_id, dtype='I'),
            'asu_id'      : rs.DataSeries(asu_id, dtype='I'),
            'image_id'    : rs.DataSeries(image_id, dtype='I'),
            'file_id'     : rs.DataSeries(file_id, dtype='I'),
            'test'        : rs.DataSeries(test_value * np.ones_like(h), dtype='I'),
        }, merged=False)
        _,idx = np.unique(output.harmonic_id, return_index=True)
        output = output.loc[idx].reset_index(drop=True)
        del(output['harmonic_id'])

        iobs = BaseModel.get_intensities(inputs).flatten()
        sig_iobs = BaseModel.get_uncertainties(inputs).flatten()
        ipred,sigipred = model.prediction_mean_stddev(inputs)
        scale,sigscale = model.scale_mean_stddev(inputs)

        num_refls = len(output)
        data_cols = {
            'Iobs'    : rs.DataSeries(iobs[:num_refls], dtype='J'),
            'SigIobs' : rs.DataSeries(sig_iobs[:num_refls], dtype='Q'),
            'Ipred'   : rs.DataSeries(ipred[:num_refls], dtype='J'),
            'SigIpred': rs.DataSeries(sigipred[:num_refls], dtype='Q'),
            'Scale'   : rs.DataSeries(scale[:num_refls], dtype='J'),
            'SigScale': rs.DataSeries(sigscale[:num_refls], dtype='Q'),
        }
        for k,v in data_cols.items():
            output[k] = v

        for i,rasu in enumerate(self.asu_collection):
            idx = output['asu_id'] == i
            result = output.loc[idx]
            result.cell = rasu.cell
            result.spacegroup = rasu.spacegroup
            yield result.set_index(['H', 'K', 'L'])


    def get_results(self, surrogate_posterior, inputs=None, output_parameters=True, max_intensity_snr=1e-5):
        """ 
        Extract results from a surrogate_posterior.

        Parameters
        ----------
        surrogate_posterior : tfd.Distribution
            A tensorflow_probability distribution or similar object with `mean` and `stddev` methods
        inputs : tuple (optional)
            Optionally use a different object from self.inputs to compute the redundancy of reflections.
        output_parameters : bool (optional)
            If True, output the parameters of the surrogate distribution in addition to the 
            moments. 
        max_intensity_snr : float (optional)
            The maximum value which will be assigned to I / SigI. 

        Returns
        -------
        results : tuple
            A tuple of rs.DataSet objects containing the results corresponding to each 
            ReciprocalASU contained in self.asu_collection
        """
        if inputs is None:
            inputs = self.inputs
        F = surrogate_posterior.mean().numpy()
        SigF = surrogate_posterior.stddev().numpy()
        I = SigF * SigF + F * F
        # <I^2> = <F^4>
        f4 = surrogate_posterior.moment_4(method='scipy')
        # var(I) = <I^2> - <I>^2 
        #        = <F^4> - <I>^2
        ivar = np.square(I * max_intensity_snr)
        ivar = np.maximum(ivar, f4 - I * I)
        SigI = np.sqrt(ivar)

        params = None
        if output_parameters:
            params = {}
            for k in sorted(surrogate_posterior.parameter_properties()):
                v = surrogate_posterior.parameters[k]
                numpify = lambda x : tf.convert_to_tensor(x).numpy()
                params[k] = numpify(v).flatten() * np.ones(len(F), dtype='float32')
        asu_id,H = self.asu_collection.to_asu_id_and_miller_index(np.arange(len(F)))
        h,k,l = H.T
        refl_id = BaseModel.get_refl_id(inputs)
        N = np.bincount(refl_id.flatten(), minlength=len(F)).astype('float32')
        results = ()
        for i,asu in enumerate(self.asu_collection):
            multiplicity = asu.multiplicity.astype('float32')
            idx = asu_id == i
            idx = idx.flatten()
            output = rs.DataSet({
                'H' : h[idx],
                'K' : k[idx],
                'L' : l[idx],
                'F' : F[idx],
                'SigF' : SigF[idx],
                'I' : I[idx],
                'SigI' : SigI[idx],
                'N' : N[idx],
                }, 
                cell=asu.cell, 
                spacegroup=asu.spacegroup,
                merged=True,
            ).infer_mtz_dtypes().set_index(['H', 'K', 'L'])
            if params is not None:
                for key in sorted(params.keys()):
                    val = params[key]
                    output[key] = rs.DataSeries(val[idx], index=output.index, dtype='R')

            # Remove unobserved refls
            output = output[output.N > 0] 

            # Reformat anomalous data
            if asu.anomalous:
                output = output.unstack_anomalous()
                # PHENIX will expect the sf / error keys in a particular order.
                anom_keys = [
                    'F(+)', 'SigF(+)', 'F(-)', 'SigF(-)', 
                    'I(+)', 'SigI(+)', 'I(-)', 'SigI(-)', 
                    'N(+)', 'N(-)'
                ]
                reorder = anom_keys + [key for key in output if key not in anom_keys]
                output = output[reorder]

            results += (output, )
        return results

    # <-- start xval data splitting methods
    def split_mono_data_by_mask(self, test_idx):
        """
        Method for splitting mono data given a boolean mask. 

        Parameters
        ----------
        test_idx : array (boolean)
            Boolean array with length of inputs.

        Returns
        -------
        train : tuple
        test  : tuple
        """
        test,train = (),()
        for inp in self.inputs:
            test  += (inp[ test_idx.flatten(),...] ,)
            train += (inp[~test_idx.flatten(),...] ,)
        return train, test

    def split_data_by_refl(self, test_fraction=0.5):
        """
        Method for splitting data given a boolean mask. 

        Parameters
        ----------
        test_fraction : float (optional)
            The fraction of reflections which will be reserved for testing.

        Returns
        -------
        train : tuple
        test  : tuple
        """
        if BaseModel.is_laue(self.inputs):
            harmonic_id = BaseModel.get_harmonic_id(self.inputs)
            test_idx = (np.random.random(harmonic_id.max()+1) <= test_fraction)[harmonic_id]
            train, test = self.split_laue_data_by_mask(test_idx)
            #return self.get_tf_dataset(train), self.get_tf_dataset(test)
            return train, test

        test_idx = np.random.random(len(self.inputs[0])) <= test_fraction
        train, test = self.split_mono_data_by_mask(test_idx)
        #return self.get_tf_dataset(train), self.get_tf_dataset(test)
        return train, test

    def split_laue_data_by_mask(self, test_idx):
        """
        Method for splitting laue data given a boolean mask. 
        This method will split up the data and alter the harmonic_id
        column to reflect the decrease in size of the array. 

        Parameters
        ----------
        test_idx : array (boolean)
            Boolean array with length of inputs.

        Returns
        -------
        train : tuple
        test  : tuple
        """
        harmonic_id = BaseModel.get_harmonic_id(self.inputs)

        # Let us just test that the boolean mask is valid for these data.
        # If it does not split observations, isect should be empty
        isect = np.intersect1d(
            harmonic_id[test_idx].flatten(),
            harmonic_id[~test_idx].flatten(),
        )
        if len(isect) > 0:
            raise ValueError(f"test_idx splits harmonic observations with harmonic_id : {isect}")

        def split(inputs, idx):
            harmonic_id = BaseModel.get_harmonic_id(inputs)

            result = ()
            uni,inv = np.unique(harmonic_id[idx], return_inverse=True)
            for i,v in enumerate(inputs):
                name = BaseModel.get_name_by_index(i)
                if name in ('intensities', 'uncertainties'):
                    v = v[uni]
                    v = np.pad(v, [[0, len(inv) - len(v)], [0, 0]], constant_values=1.)
                elif name == 'harmonic_id':
                    v = inv[:,None]
                else:
                    v = v[idx.flatten(),...]
                result += (v ,)
            return result

        return split(self.inputs, ~test_idx), split(self.inputs, test_idx)

    def split_data_by_image(self, test_fraction=0.5):
        """
        Method for splitting data given a boolean mask. 
        This method will designate full images as belonging to the 
        train or test sets. 

        Parameters
        ----------
        test_fraction : float (optional)
            The fraction of images which will be reserved for testing.

        Returns
        -------
        train : tuple
        test  : tuple
        """
        image_id = BaseModel.get_image_id(self.inputs)
        test_idx = np.random.random(image_id.max()+1) <= test_fraction

        # Low image count edge case (mostly just for testing purposes)
        if True not in test_idx:
            test_idx[0] = True
        elif False not in test_idx:
            test_idx[0] = False
            
        test_idx = test_idx[image_id]
        if BaseModel.is_laue(self.inputs):
            train, test = self.split_laue_data_by_mask(test_idx)
        else:
            train, test = self.split_mono_data_by_mask(test_idx)

        #return self.get_tf_dataset(train), self.get_tf_dataset(test)
        return train, test
    # --> end xval data splitting methods

    def build_model(self, parser=None, surrogate_posterior=None, prior=None, likelihood=None, scaling_model=None, mc_sample_size=None):
        """
        Build the model specified in parser, a careless.parser.parser.parse_args() result. Optionally override any of the 
        parameters taken by the VariationalMergingModel constructor.
        The `parser` parameter is required if self.parser is not set. 
        """
        from careless.models.merging.surrogate_posteriors import TruncatedNormal
        from careless.models.merging.variational import VariationalMergingModel
        from careless.models.scaling.image import HybridImageScaler,ImageScaler
        from careless.models.scaling.nn import MLPScaler
        if parser is None:
            parser = self.parser
        if parser is None:
            raise ValueError("No parser supplied, but self.parser is unset")

        if parser.type == 'poly':
            if parser.refine_uncertainties:
                from careless.models.likelihoods.laue import NormalEv11Likelihood as NormalLikelihood
                from careless.models.likelihoods.laue import StudentTEv11Likelihood as StudentTLikelihood
            else:
                from careless.models.likelihoods.laue import NormalLikelihood,StudentTLikelihood
        elif parser.type == 'mono':
            if parser.refine_uncertainties:
                from careless.models.likelihoods.mono import NormalEv11Likelihood as NormalLikelihood
                from careless.models.likelihoods.mono import StudentTEv11Likelihood as StudentTLikelihood
            else:
                from careless.models.likelihoods.mono import NormalLikelihood,StudentTLikelihood

        parents = parser.parents
        r_values = parser.dwr
        if prior is None and parents is None:
            prior = self.get_wilson_prior(parser.wilson_prior_b)
        elif prior is None and parser.parents is not None:
            parents = [None if i == 'None' else int(i) for i in parents.split(',')]
            r_values = [float(i) for i in r_values.split(',')]
            for r in r_values:
                if (r >= 1.) or (r <= -1.):
                    raise ValueError(
                        f"Supplied --double-wilson-r value {r} outside of allowed range (-1, 1)"
                    )
                if r < 0:
                    from warnings import warn
                    warn(f"Supplied --double-wilson-r value {r} is negative")

            sigma = self.get_wilson_sigma(parser.wilson_prior_b)
            reindexing_ops = parser.reindexing_ops
            if reindexing_ops is not None:
                delim = ';'
                reindexing_ops = [gemmi.Op(i) for i in reindexing_ops.split(delim)]

            prior = DoubleWilsonPrior(self.asu_collection, parents, r_values, reindexing_ops, sigma=sigma)

        loc,scale = prior.mean(),prior.stddev()
        scale = scale * parser.structure_factor_init_scale
        low = (1e-32 * ~self.asu_collection.centric).astype('float32')
        if surrogate_posterior is None:
            surrogate_posterior = TruncatedNormal.from_loc_and_scale(loc, scale, low, scale_shift=parser.epsilon)

        if likelihood is None:
            dof = parser.studentt_likelihood_dof
            if dof is None:
                likelihood = NormalLikelihood()
            else:
                likelihood = StudentTLikelihood(dof)

        if scaling_model is None:
            mlp_width = parser.mlp_width
            if mlp_width is None:
                mlp_width = BaseModel.get_metadata(self.inputs).shape[-1]

            if parser.scale_bijector.lower() == 'softplus':
                from tensorflow_probability import bijectors as tfb
                scale_bijector = tfb.Chain([
                    tfb.Shift(parser.epsilon),
                    tfb.Softplus(),
                ])
            elif parser.scale_bijector.lower() == 'exp':
                from tensorflow_probability import bijectors as tfb
                scale_bijector = tfb.Chain([
                    tfb.Shift(parser.epsilon),
                    tfb.Exp(),
                ])
            else:
                raise ValueError(f"Unsupported scale bijector type, {parser.scale_bijector}")

            istd = BaseModel.get_intensities(self.inputs).std()

            if parser.image_layers > 0:
                from careless.models.scaling.image import NeuralImageScaler
                n_images = np.max(BaseModel.get_image_id(self.inputs)) + 1
                scaling_model = NeuralImageScaler(
                    parser.image_layers,
                    n_images,
                    parser.mlp_layers,
                    mlp_width,
                    epsilon=parser.epsilon,
                    scale_bijector=scale_bijector,
                    scale_multiplier=istd,
                )
            else:
                mlp_scaler = MLPScaler(
                    parser.mlp_layers, mlp_width, 
                    epsilon=parser.epsilon, scale_bijector=scale_bijector, scale_multiplier=istd,
                )
                if parser.use_image_scales:
                    n_images = np.max(BaseModel.get_image_id(self.inputs)) + 1
                    image_scaler = ImageScaler(n_images)
                    scaling_model = HybridImageScaler(mlp_scaler, image_scaler)
                else:
                    scaling_model = mlp_scaler

        from tensorflow_probability import distributions as tfd
        model = VariationalMergingModel(surrogate_posterior, prior, likelihood, scaling_model, parser.mc_samples, kl_weight=parser.kl_weight)

        opt = tfk.optimizers.Adam(
            parser.learning_rate,
            parser.beta_1,
            parser.beta_2,
            clipnorm=parser.clipnorm,
            clipvalue=parser.clipvalue,
            global_clipnorm=parser.global_clipnorm,
        )

        model.compile(
            opt,
            run_eagerly=parser.run_eagerly,
        )
        return model



import tensorflow as tf

def disable_gpu():
    # Disable all GPUS
    tf.config.set_visible_devices([], 'GPU')
    visible_devices = tf.config.get_visible_devices()
    status = True
    for device in visible_devices:
        status &= device.device_type != 'GPU'
    return status




class Prior():
    """Base class for prior distributions on merged normalized structure factor amplitudes."""
    def log_prob(self):
        raise NotImplementedError("No log_prob method defined. All Priors must implement a log_prob method")





name = "Positional Encoding"
description = """
The positional encoding is a data augmentation strategy first developed in the natural language processing literature. 
It has gained widespread adoption as a simple and effective strategy to improve the performance of neural networks in 
modelling spatially resolved functions (https://arxiv.org/abs/2003.08934, https://arxiv.org/abs/1909.05215). Careless
provides the ability to apply the positional encoding to a subset of reflection metadata using the 
`--positional-encoding-keys` flag. \n

  Examples\n
  --------\n
    careless mono --positional-encoding-keys="XDET,YDET" "Hobs,Kobs,Lobs,BATCH" input.mtz out\n
"""

args_and_kwargs = (

    (("--positional-encoding-keys", ), {
        "help":"If the `--positional-encoding-frequencies` flag is set to an integer > 1, this parameter enables encoding a specific subset of"
               'of mtz columns. Supply a comma separated string of metadata keys (ie "XDET,YDET"), and these keys will be encoded separately and '
               'appended to the rest of the metadata. ', 
        "type" : str,
        "default" : None,
    }),

    (("--positional-encoding-frequencies", "-L"), {
        "help":"Number of positional encoding frequencies to apply to metadata. "
               "If you use this option, it should be paired with 'mlp-width=' in order to prevent the model from using too much memory."
               "By default all metadata columns will be encoded using the same formula. To encode a subset of the columns, please see"
               "the `--positional-encoding-keys` parameter",
        "type" : int,
        "default" : 4,
    }),

)



"""
Compute CCpred from careless output.

"""
import argparse
import numpy as np
import reciprocalspaceship as rs
import gemmi

import matplotlib.pyplot as plt
import seaborn as sns


class BaseParser(argparse.ArgumentParser):
    def __init__(self, **kwargs):
        super().__init__(
            formatter_class=argparse.RawTextHelpFormatter, 
            **kwargs
        )

        self.add_argument(
            "-s",
            "--show",
            action="store_true",
            help="Make a plot of the results and display it using matplotlib.",
        )

        self.add_argument(
            "-i",
            "--image",
            type=str,
            default=None,
            help="Make a plot of the results and save it to this filename. "
                 "The filetype will be determined from the filename. "
                 "Any filetype supported by your matplotlib version will be available.",
        )

        self.add_argument(
            "-o",
            "--output",
            type=str,
            default=None,
            help="Optionally save results to this file in csv format instead of printing "
                 "them to the terminal.",
        )

        self.add_argument(
            "-l",
            "--ylim",
            type=float,
            nargs=2,
            default=None,
            help="Override the y-axis limits like `--ylim 0. 1.`"
        )




import reciprocalspaceship as rs
import pandas as pd
import numpy as np

class ReciprocalASU():
    def __init__(self, cell, spacegroup, dmin, anomalous):
        """
        Parameters
        ----------
        cell : gemmi.UnitCell
            A gemmi.UnitCell instance
        spacegroup : gemmi.SpaceGroup
            A gemmi.SpaceGroup instance
        dmin : float
            The maximum resolution in Angstroms
        anomalous : bool
            Whether the ASU includes Friedel minus acentric reflections
        """
        self.cell = cell
        self.spacegroup = spacegroup
        self.dmin = dmin
        self.anomalous = anomalous
        self.Hall = rs.utils.generate_reciprocal_asu(
                self.cell, 
                self.spacegroup, 
                self.dmin, 
                self.anomalous
            )
        h,k,l = self.Hall.T
        lookup_table = rs.DataSet({
            'H' : h,
            'K' : k,
            'L' : l,
            'id' : np.arange(len(h))
            }, 
            cell=cell,
            spacegroup=spacegroup,
        ).compute_multiplicity().label_centrics().compute_dHKL()
        self.lookup_table = lookup_table

    @property
    def centric(self):
        """ boolean array true for centric refl_ids """
        return self.lookup_table.CENTRIC.to_numpy('bool')

    @property
    def multiplicity(self):
        """ the multiplicity of each structure factor """
        return self.lookup_table.EPSILON.to_numpy('float32')

    @property
    def dHKL(self):
        """ the multiplicity of each structure factor """
        return self.lookup_table.dHKL.to_numpy('float32')

    def to_refl_id(self, H):
        """
        Parameters
        ----------
        H : np.array
            (n x 3) array of miller indices to convert to integer reflection ids

        Returns
        -------
        refl_id : np.array
            (n x 1) array of integer reflection ids
        """
        idx = pd.MultiIndex.from_arrays(H.T, names = ['H', 'K', 'L'])
        return self.lookup_table.set_index(['H','K','L']).loc[idx, 'id']

    def to_miller_index(self, refl_id):
        """
        Parameters
        ----------
        refl_id : np.array
            (n x 1) array of integer reflection ids to convert to miller indices

        Returns
        -------
        H : np.array
            (n x 3) array of miller indices 
        """
        return self.lookup_table.set_index('id').loc[refl_id].get_hkls()

class ReciprocalASUCollection():
    def __init__(self, reciprocal_asus):
        """
        Parameters
        ----------
        reciprocal_asus : list
            A list of ReciprocalAsu instances
        """
        self.reciprocal_asus = reciprocal_asus
        self.lookup_table = None
        for asu_id,asu in enumerate(self.reciprocal_asus):
            tab = asu.lookup_table.copy()
            tab['asu_id'] = asu_id
            if self.lookup_table is not None:
                tab['id'] += len(self.lookup_table)
                self.lookup_table = rs.concat((self.lookup_table, tab), check_isomorphous=False)
            else:
                self.lookup_table =  tab
        self.asu_and_miller_lookup_table = self.lookup_table.set_index(['asu_id', 'H', 'K', 'L'])
        self.refl_id_lookup_table = self.lookup_table.set_index('id')

    @property
    def centric(self):
        """ boolean array true for centric refl_ids """
        return self.refl_id_lookup_table.CENTRIC.to_numpy('bool')

    @property
    def multiplicity(self):
        """ the multiplicity of each structure factor """
        return self.refl_id_lookup_table.EPSILON.to_numpy('float32')

    @property
    def dHKL(self):
        """ the multiplicity of each structure factor """
        return self.lookup_table.dHKL.to_numpy('float32')

    @property
    def hkls(self):
        """ the multiplicity of each structure factor """
        return self.refl_id_lookup_table.get_hkls()

    @property
    def asu_ids(self):
        """ the multiplicity of each structure factor """
        return self.refl_id_lookup_table.asu_id.to_numpy('int')

    def to_asu_id_and_miller_index(self, refl_id):
        """
        Parameters
        ----------
        refl_id : np.array
            (n x 1) array of integer reflection ids to convert to miller indices

        Returns
        -------
        asu_id : np.array
            (n x 1) array of asu ids 
        H : np.array
            (n x 3) array of miller indices 
        """
        data = self.refl_id_lookup_table.loc[refl_id.flatten()]
        H = data.get_hkls()
        asu_id = data['asu_id'].to_numpy('int')
        return asu_id[:,None], H

    def to_refl_id(self, asu_id, H, allow_missing=False):
        """
        Parameters
        ----------
        asu_id : np.array
            (n x 1) array of asu ids 
        H : np.array
            (n x 3) array of miller indices 
        allow_missing : bool (optional)
            Replace missing reflections with -1 rather than throwing an error

        Returns
        -------
        refl_id : np.array
            (n x 1) array of integer reflection ids to convert to miller indices
        """
        idx = np.concatenate((asu_id, H), axis=1).astype('int')
        idx = pd.MultiIndex.from_arrays(idx.T, names = ['asu_id', 'H', 'K', 'L'])
        if allow_missing:
            refl_ids = pd.DataFrame(index=idx).join(self.asu_and_miller_lookup_table.id)
            refl_ids = refl_ids.fillna(-1).to_numpy("int").flatten()
            return refl_ids
        return self.asu_and_miller_lookup_table.loc[idx, 'id'].to_numpy('int').flatten()

    def __getitem__(self, i):
        return self.reciprocal_asus[i]

    def __len__(self):
        return len(self.reciprocal_asus)




from careless.models.likelihoods.mono import Likelihood
from tensorflow_probability import distributions as tfd
from tensorflow_probability import bijectors as tfb
import tensorflow_probability as tfp
import tensorflow as tf
import math
import numpy as np

class ConvolvedLikelihood():
    """
    Convolved log probability object for Laue data.
    """
    def __init__(self, distribution, harmonic_id):
        self.harmonic_id = harmonic_id
        self.distribution = distribution

    def convolve(self, value):
        """
        Takes a set of sample points at which to compute the log prob. 
        values can either be a bare vector or it may have a batch
        dimension for mc samples, ie shape=(b, n_predictions). 
        """
        tv = tf.transpose(value)
        tr = tf.scatter_nd(self.harmonic_id, tv, tv.shape)
        return tf.transpose(tr)

    def mean(self, *args, **kwargs):
        return self.distribution.mean(*args, **kwargs)

    def stddev(self, *args, **kwargs):
        return self.distribution.stddev(*args, **kwargs)

    def log_prob(self, value):
        return self.distribution.log_prob(self.convolve(value))

class LaueBase(Likelihood):
    def dist(self, inputs):
        raise NotImplementedError(
            """ Extensions of this class must implement self.location_scale_distribution(loc, scale) """
            )

    def call(self, inputs):
        harmonic_id   = self.get_harmonic_id(inputs)

        likelihood = self.dist(inputs)

        return ConvolvedLikelihood(likelihood, harmonic_id)

from careless.models.likelihoods.mono import NormalEv11Likelihood as MonoNormalEv11Likelihood
class NormalEv11Likelihood(LaueBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.mono = MonoNormalEv11Likelihood()

    def dist(self, inputs):
        return self.mono(inputs)

from careless.models.likelihoods.mono import StudentTEv11Likelihood as MonoStudentTEv11Likelihood
class StudentTEv11Likelihood(LaueBase):
    def __init__(self, dof, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.mono = MonoStudentTEv11Likelihood(dof)

    def dist(self, inputs):
        return self.mono(inputs)


class NormalLikelihood(LaueBase):
    def dist(self, inputs):
        loc = self.get_intensities(inputs)
        scale = self.get_uncertainties(inputs)
        loc = tf.squeeze(loc)
        scale = tf.squeeze(scale)
        return tfd.Normal(loc, scale)

class LaplaceLikelihood(LaueBase):
    def dist(self, inputs):
        loc = self.get_intensities(inputs)
        scale = self.get_uncertainties(inputs)
        loc = tf.squeeze(loc)
        scale = tf.squeeze(scale)
        return tfd.Laplace(loc, scale / math.sqrt(2.))

class StudentTLikelihood(LaueBase):
    def __init__(self, dof):
        """
        Parameters
        ----------
        dof : float
            Degrees of freedom of the t-distributed error model.
        """
        super().__init__()
        self.dof = dof

    def dist(self, inputs):
        loc = self.get_intensities(inputs)
        scale = self.get_uncertainties(inputs)
        loc = tf.squeeze(loc)
        scale = tf.squeeze(scale)
        return tfd.StudentT(self.dof, loc, scale)




"""
Compute CCanom from careless output.
"""
import argparse
import numpy as np
import matplotlib.pyplot as plt
import reciprocalspaceship as rs
import seaborn as sns


from careless.stats.parser import BaseParser
class ArgumentParser(BaseParser):
    def __init__(self):
        super().__init__(
            description=__doc__
        )

        # Required arguments
        self.add_argument(
            "mtz",
            nargs="+",
            help="MTZs containing crossvalidation data from careless",
        )

        # Optional arguments
        self.add_argument(
            "-m",
            "--method",
            default="weighted",
            choices=["weighted", "pearson", "spearman"],
            help="Method for computing correlation coefficient (weighted, spearman, or pearson). "
            "Weighted is the default which is consistent with the behavior of XDS`.",
        )

        self.add_argument(
            "-b",
            "--bins",
            default=10,
            type=int,
            help="Number of resolution bins to use, the default is 10.",
        )

        self.add_argument(
            '--overall',
            action="store_true",
            help="Pool all prediction mtz files into a single calculation rather than treating each file individually.",
        )

def make_halves_ccanom(mtz, bins=10):
    """Construct half-datasets for computing CCanom"""
    #Only user rows with both F+ and F-
    mtz = mtz.loc[(mtz["N(+)"] > 0) & (mtz["N(-)"] > 0)]

    half1 = mtz.loc[mtz.half == 0].copy()
    half2 = mtz.loc[mtz.half == 1].copy()

    half1["Danom"] = half1["F(+)"] - half1["F(-)"]
    half2["Danom"] = half2["F(+)"] - half2["F(-)"]
    half1["SigDanom"] = np.sqrt(
        np.square(half1["SigF(+)"]) + np.square(half1["SigF(-)"])
    )
    half2["SigDanom"] = np.sqrt(
        np.square(half2["SigF(+)"]) + np.square(half2["SigF(-)"])
    )

    out = half1[["Danom", "SigDanom", "repeat"]].merge(
        half2[["Danom", "SigDanom", "repeat"]], on=["H", "K", "L", "repeat"], suffixes=("1", "2")
    )
    return out

def weighted_pearson_ccfunc(df):
    x = df['Danom1'].to_numpy('float32')
    y = df['Danom2'].to_numpy('float32')
    w = np.reciprocal(
        np.square(df['SigDanom1']) + np.square(df['SigDanom2'])
    ).to_numpy('float32')
    return rs.utils.weighted_pearsonr(x, y, w)

def spearman_ccfunc(df):
    return df[['Danom1', 'Danom2']].corr(method='spearman')['Danom1']['Danom2']

def pearson_ccfunc(df):
    return df[['Danom1', 'Danom2']].corr(method='pearson')['Danom1']['Danom2']

def run_analysis(args):
    ds = []
    for m in args.mtz:
        _ds = rs.read_mtz(m)
        #non-isomorphism could lead to different resolution for each mtz
        #need to calculate dHKL before concatenating 
        _ds = make_halves_ccanom(_ds)
        _ds.compute_dHKL(inplace=True)
        _ds['file'] = m
        _ds['Spacegroup'] = _ds.spacegroup.xhm()
        ds.append(_ds)
    ds = rs.concat(ds, check_isomorphous=False)
    bins,edges = rs.utils.bin_by_percentile(ds.dHKL, args.bins, ascending=False)
    ds['bin'] = bins
    labels = [
        f"{e1:0.2f} - {e2:0.2f}"
        for e1, e2 in zip(edges[:-1], edges[1:])
    ]

    if args.overall:
        grouper = ds.groupby(["bin", "repeat"])
    else:
        grouper = ds.groupby(["file", "bin", "repeat"])
    if args.method.lower() == "spearman":
        ccfunc = spearman_ccfunc
    elif args.method.lower() == "pearson":
        ccfunc = pearson_ccfunc
    elif args.method.lower() == "weighted":
        ccfunc = weighted_pearson_ccfunc
    else:
        raise ValueError(f"Unrecognized CC --method, {args.method}")

    result = grouper.apply(ccfunc)
    result = rs.DataSet({"CCanom" : result}).reset_index()
    result['Resolution Range (Å)'] = np.array(labels)[result.bin]
    result['Spacegroup'] = grouper['Spacegroup'].apply('first').to_numpy()
    if not args.overall:
        result['file'] = grouper['file'].apply('first').to_numpy()
        result = result[['file', 'repeat', 'Resolution Range (Å)', 'bin', 'Spacegroup', 'CCanom']]
    else:
        result = result[['repeat', 'Resolution Range (Å)', 'bin', 'Spacegroup', 'CCanom']]

    if args.output is not None:
        result.to_csv(args.output)
    else:
        print(result.to_string())

    plot_kwargs = {
        'data' : result,
        'x' : 'bin',
        'y' : 'CCanom',
    }

    if args.overall:
        plot_kwargs['color'] = 'k'
    else:
        plot_kwargs['hue'] = 'file'
        plot_kwargs['palette'] = "Dark2"

    sns.lineplot(**plot_kwargs)
 
    plt.xticks(range(args.bins), labels, rotation=45, ha="right", rotation_mode="anchor")
    plt.ylabel(r"$\mathrm{CC_{anom}}$ " + f"({args.method})")
    plt.xlabel("Resolution ($\mathrm{\AA}$)")
    plt.grid(which='both', axis='both', ls='dashdot')
    if args.ylim is not None:
        plt.ylim(args.ylim)
    plt.tight_layout()

    if args.image is not None:
        plt.savefig(args.image)

    if args.show:
        plt.show()

def main():
    parser = ArgumentParser().parse_args()
    run_analysis(parser)




# Special groups
from . import required
from . import poly

# Groups
from . import (
    common,
    crossvalidation,
    filtration,
    interpretation,
    likelihood,
    optimizer,
    positional_encoding,
    prior,
    scaling,
    tf_options,
)

groups = [
    common,
    crossvalidation,
    filtration,
    interpretation,
    likelihood,
    optimizer,
    positional_encoding,
    prior,
    scaling,
    tf_options
]




import numpy as np
import reciprocalspaceship as rs


def calculate_harmonic(H):
    n = np.gcd.reduce(H, axis=-1)
    return n

@rs.decorators.range_indexed
def expand_harmonics(ds, dmin=None,  wavelength_key='Wavelength'):
    """
    Expand reflection observations to include all contributing harmonics. All 
    contributing reflections will be included out to a resolution cutoff 
    irrespective of peak wavelength.

    Parameters
    ----------
    ds : rs.DataSet
        Laue data without multiples. These should be unmerged data with 
        miller indices that have not been mapped to the ASU (P1 Miller Indices).
    dmin : float
        Highest resolution in Å to which harmonics will be predicted. If not 
        supplied, the highest resolution reflection in ds will set dmin.

    Returns
    -------
    ds : rs.DataSet
        DataSet with all reflection observations expanded to include their 
        constituent reflections. New columns 'H_0', 'K_0', 'L_0' will be added 
        to each reflection to store the Miller indices of the innermost 
        reflection on each central ray. 
    """
    if ds.merged:
        raise ValueError("Expected unmerged data, but ds.merged is True")

    ds = ds.copy()

    #Here's where we get the metadata for Laue harmonic deconvolution
    #This is the HKL of the closest refl on each central ray
    if 'dHKL' not in ds:
        ds.compute_dHKL(inplace=True)
    if dmin is None:
        dmin = ds['dHKL'].min() - 1e-12

    Hobs = ds.get_hkls()

    #Calculated the harmonic as indexed
    nobs = calculate_harmonic(Hobs)

    #Add primary harmonic miller index, wavelength, and resolution
    # H = H_n / n
    # lambda = lambda_n * n
    # d = d_n * n
    H_0 = (Hobs/nobs[:,None]).astype(np.int32)
    d_0 = ds['dHKL'].to_numpy() * nobs
    Wavelength_0 = ds[wavelength_key].to_numpy() * nobs

    #This is the largest harmonic that should be
    #included for each observation in order to
    #respect the resolution cutoff
    n_max =  np.floor_divide(d_0, dmin).astype(int)

    #This is where we make the indices to expand
    #each harmonic the appropriate number of times given dmin
    n = np.arange(n_max.max()) + 1
    idx,n = np.where(n[None,:] <= n_max[:,None])
    n = n + 1
    #idx are the indices for expansion and n is the corresponding
    #set of harmonic integers

    ds = ds.iloc[idx]
    ds.reset_index(inplace=True, drop=True)
    ds['H_0'],ds['K_0'],ds['L_0'] = H_0[idx].T
    ds[wavelength_key] = (Wavelength_0[idx] / n)
    ds['H'],ds['K'],ds['L'] = (n[:,None] * H_0[idx]).T

    #Update dHKL using the cell and not harmonics
    #d_0[idx] / n is not numerically precise enough
    ds.compute_dHKL(inplace=True)

    return ds



from careless.models.likelihoods.base import Likelihood
import tf_keras as tfk
import tensorflow as tf
from tensorflow_probability import distributions as tfd
from tensorflow_probability import util as tfu
from tensorflow_probability import bijectors as tfb
import math
import numpy as np

class LocationScaleLikelihood(Likelihood):
    def get_loc_and_scale(self, inputs):
        loc   = self.get_intensities(inputs)
        scale = self.get_uncertainties(inputs)
        return tf.squeeze(loc), tf.squeeze(scale)

class NormalLikelihood(LocationScaleLikelihood):
    def call(self, inputs):
        return tfd.Normal(*self.get_loc_and_scale(inputs))

class LaplaceLikelihood(LocationScaleLikelihood):
    def call(self, inputs):
        loc, scale = self.get_loc_and_scale(inputs)
        return tfd.Laplace(loc, scale/math.sqrt(2.))

class StudentTLikelihood(LocationScaleLikelihood):
    def __init__(self, dof):
        """
        Parameters
        ----------
        dof : float
            Degrees of freedom of the student t likelihood.
        """
        super().__init__()
        self.dof = dof

    def call(self, inputs):
        return tfd.StudentT(self.dof, *self.get_loc_and_scale(inputs))

class Ev11Likelihood(LocationScaleLikelihood):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.Sdfac = tfu.TransformedVariable(1., tfb.Softplus())
        self.Sdadd = tfu.TransformedVariable(1., tfb.Softplus())
        self.SdB = tfu.TransformedVariable(1., tfb.Softplus())
        self.loc = None
        self.scale = None

    def call(self, inputs):
        self.loc, self.scale = self.get_loc_and_scale(inputs)
        return self

    def corrected_sigiobs(self, ipred):
        ipred = tf.math.softplus(ipred)
        sigiobs = self.Sdfac * tf.math.sqrt(
            tf.square(self.scale) + \
            self.SdB * ipred + \
            self.Sdadd * tf.square(ipred)
        )
        return sigiobs

class NormalEv11Likelihood(Ev11Likelihood):
    def log_prob(self, ipred):
        scale = self.corrected_sigiobs(ipred)
        return tfd.Normal(self.loc, scale).log_prob(ipred)

class StudentTEv11Likelihood(Ev11Likelihood):
    def __init__(self, dof, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.dof = dof

    def log_prob(self, ipred):
        scale = self.corrected_sigiobs(ipred)
        return tfd.StudentT(self.dof, self.loc, scale).log_prob(ipred)

class NeuralLikelihood(Likelihood):
    def __init__(self, mlp_layers, mlp_width):
        super().__init__()

        layers = []
        for _ in range(mlp_layers):
            layer = tfk.layers.Dense(
                mlp_width,
                activation=tfk.layers.LeakyReLU(),
            )
            layers.append(layer)

        layer = tfk.layers.Dense(
            1,
            activation='softplus',
        )
        layers.append(layer)
        self.network = tfk.models.Sequential(layers)

    def base_dist(self, loc, scale):
        raise NotImplementedError("extensions of this class must implement a base_dist(loc, scale) method")

    def call(self, inputs):
        iobs = self.get_intensities(inputs)
        #metadata = self.get_metadata(inputs)
        sigiobs = self.get_uncertainties(inputs)
        delta = self.network(tf.concat((iobs, sigiobs), axis=-1))
        sigpred = sigiobs * delta / tf.reduce_mean(delta)
        return self.base_dist(
            tf.squeeze(iobs), 
            tf.squeeze(sigpred),
        )

class NeuralNormalLikelihood(NeuralLikelihood):
    def base_dist(self, loc, scale):
        return tfd.Normal(loc, scale)



name = "TensorFlow"
description = None

args_and_kwargs = (
    (("--run-eagerly",), {
        "help":"Running tensorflow in eager mode may be required for high memory models.", 
        "action":'store_true', 
        "default":False,
    }),

    (("--jit-compile",), {
        "help":"Use jit compilation. By default defer to TF to decide. Enabling this option may decrease memory requirements for some models at the expense of increasing startup time.", 
        "action":'store_true', 
        "default":None,
    }),

    (("--reduce-retracing",), {
        "help":"Use the reduce retracing option during compilation. This is disabled by default.", 
        "action":'store_true', 
        "default": False,
    }),

    (("--disable-gpu",), {
        "help":"Disable GPU for high memory models.", 
        "action":'store_true', 
        "default":False,
    }),

    (("--gpu-id",), {
        "help":"Specify the physical device used for acceleration. This is an integer from"
               "0 to num accelerators - 1. The default is zero. If `--disable-gpu` is set,"
               "this option is ignored.",
        "type":int,
        "default": 0,
    }),

    (("--disable-memory-growth",), {
        "help":"Disable the experimental dynamic memory allocation.", 
        "action":'store_true', 
        "default":False,
    }),

    (("--tf-debug",), {
        "help": "Increase the TensorFlow log verbosity by setting the "
                "TF_CPP_MIN_LOG_LEVEL environment variable. ",
        "action" : 'store_true',
        "default":False,
    }),

    (("--seed",), { 
        "help":f"Random number seed for consistent sampling.", 
        "type":int, 
        "default":1234, 
    }),

)



name = "Optimizer Parameters"
description = None

args_and_kwargs = (
    (("--iterations",), {
        "help":"Number of gradient steps to take.", 
        "type":int, 
        "default":10000,
    }),

    (("--learning-rate",), {
        "help":"Adam learning rate. The default is 0.001", 
        "type":float, 
        "default":0.001,
    }),

    (("--beta-1",), {
        "help":"Adam beta_1 param. The default is 0.9", 
        "type":float, 
        "default":0.9,
    }),

    (("--beta-2",), {
        "help":"Adam beta_2 param. The default is 0.99", 
        "type":float, 
        "default":0.99,
    }),

    (("--clipnorm",), {
        "help":"Optionally clip the norm of the gradient of each weight to be no larger than this value.", 
        "type": float, 
        "default": None,
    }),

    (("--clipvalue",), {
        "help":"Optionally clip the gradients to be no larger than this value.", 
        "type": float, 
        "default": None,
    }),

    (("--global-clipnorm",), {
        "help":"Optionally clip the norm of all the gradients to be no larger than this value.", 
        "type": float, 
        "default": None,
    }),

)



name = None
description = None

args_and_kwargs = (
    # Required args
    (("metadata_keys", ),  {
        "help":"Metadata keys for scaling. This is expected to be a comma delimitted string. "
               "You may use `rs.mtzdump` to check the available metadata column names in mtz files. "
               "Careless always provides the special metadata keys, "
               "'dHKL,Hobs,Kobs,Lobs,image_id'. "
               "For stream files, careless does not support arbitrary metadata columns but rather "
               "provides the metadata keys, "
               "'BATCH,s1x,s1y,s1z,ewald_offset,angular_ewald_offset', "
               "These correspond to image number, scattered beam wavevectors (x,y,z), and ewald offsets (inverse angstroms, degrees).",
        "type":str, 
    }),

    (("reflection_files", ), { 
        "metavar":"reflections.{mtz,stream}", 
        "help":"Mtz or stream file(s) containing unmerged reflection observations. "
               "If you are supplying stream files, you must also use the --spacegroups option to supply the symmetry for merging. "
               "See the metadata_keys param for more info about stream file usage. "
               "careless poly does not currently support .stream files.",
        "type":str, 
        "nargs":'+',
    }),

    (("output_base", ), {
        "metavar":"out", 
        "help":"Output filename base.", 
        "type":str,
    }),
)




name=None
description=None

args_and_kwargs = (
    (("--embed", ),  {
        "help":"Drop to an IPython shell at the end of optimization to inpsect variables.",
        "action" : "store_true",
        "default" : False,
    }),

    (("--mc-samples",), {
        "help":"This is the number of samples to take per gradient step with default 1. " ,
        "type": int, 
        "default" : 1,
    }),

    (("--structure-factor-file",), {
        "help": "Initialize the structure factors from the ouput of a previous run. This argument should be a string beginning with the "
                "base filename used in the previous run and ending in _structure_factor.  For instance, if the previous run "
                "was called with `careless mono [...] merge/hewl`, the appropriate filename to use would be merge/hewl_structure_factor. ",
        "type": str, 
        "default" : None,
    }),


    (("--freeze-structure-factors",), {
        "help": "Do not optimize the structure factors.",
        "action": "store_true"
    }),

    (("--structure-factor-init-scale",), {
        "help":"A floating point number usually between 0 and 1. The width of the initial structure factor distribution is this times" 
               "the standard deviation of the prior distribution. The default is 1.0. ",
        "type": float, 
        "default" : 1.0,
    }),

    (("--epsilon",), {
        "help":"A small constant added to the scale parameters of variational distributions  for numerical stability. The default is 1e-7." ,
        "type": float, 
        "default" : 1e-7,
    }),

    (("--disable-metadata-standardization",), {
        "help": "By default careless will convert metadata to z-scores. This flag disables that behavior. In general, unstandardized metadata will lead to unstable optimization. However, this flag might be useful if the user wants to use their own normalization scheme." ,
        "action": "store_false", 
        "dest" : "standardize_metadata", 
    }),

    (("--disable-progress-bar", ),  {
        "help":"Disable the progress bar. This is helpful if you're logging stderr.",
        "action" : "store_true",
        "default" : False,
    }),

    (("--save-data-manager", ),  {
        "help":"Optionally save the data manager object in pickle format.",
        "action" : "store_true",
        "default" : False,
    }),
    
)



name = "Prior"
description = """
Options related to the prior distribution applied to the structure factors during merging.
"""

args_and_kwargs = (
    (
        ("--kl-weight",),
        {
            "help": "Set the weight of the kl divergence term relative to the likelihood. "
            "By default, by default this is based purely on the number of reflections.",
            "type": float,
            "default": None,
        },
    ),
    (
        ("--wilson-prior-b",),
        {
            "help": "This flag enables learning reflections on a particular Wilson scale. "
            "By default, the Wilson prior is flat across resolution bins. ",
            "type": float,
            "default": None,
        },
    ),
    (
        ("--double-wilson-r",),
        {
            "help": "For each input mtz, designate a prior correlation coefficient with its parent. "
            "Supply one float for each file separated by commas. Supply zero for each root node."
            "for example, --double-wilson-r=0.,0.9. ",
            "type": str,
            "default": None,
            "dest": "dwr",
        },
    ),
    (
        ("--double-wilson-parents",),
        {
            "help": "For each input mtz, designate a parent upon which its prior is conditioned. "
            "Supply one integer for each file separated by commas. Supply None for root nodes"
            "which follow single Wilson priors. "
            "for example, --double-wilson-parents=None,0 ",
            "type": str,
            "default": None,
            "dest": "parents",
        },
    ),
    (
        ("--double-wilson-reindexing-ops",),
        {
            "help": 'Provide semicolon-delimited reindexing operators to remap miller indices from the child ASU to the parent. For root nodes or nodes with parents in the same setting, supply the identity operator x,y,z. For example --double-wilson-reindexing-ops="x,y,z;x-y,x,z+1/2". Note that quotations should be used in some circumstances to prevent bash from interpretting the semicolons.',
            "type": str,
            "default": None,
            "dest": "reindexing_ops",
        },
    ),
)



"""
Estimate the Wilson b-factor from unmerged data.

"""
import argparse
import numpy as np
import reciprocalspaceship as rs
import pandas as  pd
from scipy.stats import linregress


class ArgumentParser(argparse.ArgumentParser):
    def __init__(self):
        super().__init__(
            formatter_class=argparse.RawTextHelpFormatter, 
            description=__doc__
        )

        # Required arguments
        self.add_argument(
            "input",
            nargs="+",
            help="MTZs or stream files containing unmerged data",
        )

        # Optional arguments
        self.add_argument(
            "-i",
            "--intensity-key",
            type=str,
            default=None,
            help="Intensity column to use. The first one will be used by default.",
        )

        self.add_argument(
            "-s",
            "--sigma-key",
            type=str,
            default=None,
            help="Sigma intensity column to use. The first one will be used by default.",
        )

        self.add_argument(
            "-b",
            "--bins",
            type=int,
            default=20,
            help="Number of bins into which to divide the data.",
        )

        resolution_group = self.add_mutually_exclusive_group()

        resolution_group.add_argument(
            "-c",
            "--isigi-cutoff",
            type=float,
            default=1.5,
            help="If this option is supplied, the script tries to estimate an appropriate resolution" 
                 "cutoff from the signal to noise in resolution bins. The default value is 1.5 (I/SigI).",
        )

        resolution_group.add_argument(
            "-d",
            "--dmin",
            type=float,
            default=None,
            help="Minimum resolution cutoff in (Å) which overrides the more automated --isigi-cutoff.",
        )

        self.add_argument(
            "-x",
            "--dmax",
            type=float,
            default=np.inf,
            help="Maximum resolution cutoff in (Å) which defaults to infinity.",
        )

        self.add_argument(
            "--plot",
            action="store_true",
            help="Make a plot of the results and display it using matplotlib.",
        )

def _make_df(dHKL, I, SigI, bins=None):
    df = pd.DataFrame({
        'dHKL' : dHKL,
        'I' : I,
        'SigI' : SigI,
    })
    if bins is not None:
        labels, edges = rs.utils.bin_by_percentile(dHKL, bins)
        df['bin'] = labels
    return df

def _truncate_data(dHKL, I, SigI, bins=20, isigi_cutoff=None, dmin=None):
    if isigi_cutoff is None and dmin is None:
        return dHKL, I, SigI
    if dmin is not None:
        idx = dHKL >= dmin
        return dHKL[idx], I[idx], SigI[idx]
    df = _make_df(dHKL, I, SigI, bins)
    df['isigi'] = df.I / df.SigI
    mu = df.groupby('bin').mean()
    mu.isigi > isigi_cutoff
    dmin = mu.dHKL[mu.isigi >= isigi_cutoff].min()
    return _truncate_data(dHKL, I, SigI, bins, None, dmin)

def estimate_b(dHKL, I, SigI, bins=20, isigi_cutoff=None, dmin=None):
    """
    Estimate the Wilson b-factor for the data. 

    Parameters
    ----------
    dHKL : array
        The resolution of each reflection in Å.
    I : array
        The intensity of each reflection.
    SigI : array
        The uncertainty of each reflection.
    bins : int (optional)
        The number of bins to use for fitting the wilson plot
    isigi_cutoff : float (optional)
        Optionally truncate the resolution by I / SigI
    dmin : float (optional)
        Optionally truncate the resolution at a specific value in Å.
        This option supersedes isigi_cutoff
    """
    dHKL, I, SigI = _truncate_data(dHKL, I, SigI, bins, isigi_cutoff, dmin)
    df = _make_df(dHKL, I, SigI, bins)
    df['VarI'] = np.square(df.SigI)
    df['inv_d2'] = np.reciprocal(np.square(df['dHKL']))
    mu = df[['bin', 'I', 'inv_d2']].groupby('bin').mean()
    x,y = mu.inv_d2, np.log(mu.I)

    keys = ('slope', 'intercept', 'r_value', 'p_value', 'stderr', 'intercept_stderr', 'x', 'y')
    result = linregress(x, y) 
    result.x = x
    result.y = y
    return result

def run_analysis(parser):
    from careless.io.formatter import get_first_key_of_dtype
    ds = []
    for i,file in enumerate(parser.input):
        if file.endswith('.mtz'):
            _ds = rs.read_mtz(file)
        elif file.endswith('.stream'):
            _ds = rs.read_crystfel(file)

        _ds['file_id'] = i
        ds.append(_ds)
    ds = rs.concat(ds)

    ikey = parser.intensity_key
    sigkey = parser.sigma_key
    if ikey is None:
        ikey = get_first_key_of_dtype(ds, 'J')
    if sigkey is None:
        sigkey = get_first_key_of_dtype(ds, 'Q')
        for key in ds:
            if ds[key].dtype == 'Q' and key.endswith(ikey):
                sigkey = key

    ds.compute_dHKL(inplace=True)

    dHKL,I,SigI = ds[['dHKL', 'I', 'SigI']].to_numpy('float32').T
    fit = estimate_b(dHKL, I, SigI, parser.bins, parser.isigi_cutoff, parser.dmin)


    title = f"Estimated Wilson b-factor: {-2. * fit.slope:0.2f} ± {2. * fit.stderr:0.2f}"
    if parser.plot:
        import matplotlib.pyplot as plt
        import seaborn as sns
        sns.regplot(x=fit.x, y=fit.y, color='k')
        plt.xlabel(r'$1 / d_{HKL}^2\ (Å^{-2})$')
        plt.ylabel(r'$\log \langle I \rangle$')
        plt.title(title)
        plt.show()
    print(title)


def main():
    parser = ArgumentParser().parse_args()
    run_analysis(parser)


if __name__=='__main__':
    main()




"""
Compute completeness from careless output.
"""
import argparse
import matplotlib.pyplot as plt
import reciprocalspaceship as rs
import seaborn as sns
import numpy as np

from careless.stats.parser import BaseParser
class ArgumentParser(BaseParser):
    def __init__(self):
        super().__init__(
            description=__doc__
        )

        # Required arguments
        self.add_argument(
            "mtz",
            help="MTZ containing merged data from careless",
        )

        self.add_argument(
            "-b",
            "--bins",
            default=10,
            type=int,
            help=("Number of resolution bins to use, the default is 10."),
        )

def run_analysis(args):
    ds = rs.read_mtz(args.mtz)
    results = rs.stats.compute_completeness(ds, bins=args.bins)
    results = results['completeness']


    #Move overall to the beginning
    results = results.iloc[np.roll(np.arange(len(results)), 1)]
    results = results.reset_index()
    xlabel = 'Resolution Range (Å)'
    results = results.rename(columns={'index' : xlabel})

    if args.output is not None:
        results.to_csv(args.output, index=False)
    else:
        print(results.to_string(index=False))

    ax = sns.lineplot(
        data=results.melt(xlabel),
        x=xlabel,
        y='value',
        hue='variable',
        palette='Dark2',
    )
    plt.xticks(rotation=45, rotation_mode='anchor', ha='right')
    plt.legend(title="")
    plt.ylabel(r"Completeness")
    plt.grid(which='both', axis='both', ls='dashdot')
    if args.ylim is not None:
        plt.ylim(args.ylim)
    plt.tight_layout()

    if args.image is not None:
        plt.savefig(args.image)

    if args.show:
        plt.show()

def main():
    parser = ArgumentParser().parse_args()
    run_analysis(parser)




from careless.utils.distributions import Rice,FoldedNormal
from tensorflow_probability.python.internal.special_math import ndtr
import tensorflow_probability as tfp
from tensorflow_probability import distributions as tfd
from tensorflow_probability import bijectors as tfb
from tensorflow_probability.python.internal import tensor_util
import tensorflow as tf
import tf_keras as tfk
import numpy as np

class SurrogatePosterior(tfk.models.Model):
    """ The base class for learnable variational distributions over structure factor amplitudes. """
    def __init__(self, distribution, **kwargs):
        super().__init__(**kwargs)
        self.distribution = distribution

    def sample(self, *args, **kwargs):
        return self.distribution.sample(*args, **kwargs)

    def log_prob(self, *args, **kwargs):
        return self.distribution.log_prob(*args, **kwargs)

    def mean(self, *args, **kwargs):
        return self.distribution.mean(*args, **kwargs)

    def stddev(self, *args, **kwargs):
        return self.distribution.stddev(*args, **kwargs)

    def parameter_properties(self, *args, **kwargs):
        return self.distribution.parameter_properties(*args, **kwargs)

    def moment_4(self):
        raise NotImplementedError("The fourth moment of this distribution is not implemented yet.")

    @property
    def parameters(self):
        return self.distribution.parameters


#This is a temporary workaround for tfd.TruncatedNormal which has a bug in sampling
#2020-10-30: This should be removed if this issue is fixed: https://github.com/tensorflow/probability/issues/1149
#
#2020-11-01: On second thought, this may not be fixed unless they git rid of the current rejection sampler based 
# implementation. See https://github.com/tensorflow/probability/issues/518, for additional issues.
class TruncatedNormal(SurrogatePosterior):
    def __init__(self, loc, scale, low, high, validate_args=False, allow_nan_stats=True, name='TruncatedNormal', **kwargs):
        distribution = tfd.TruncatedNormal(loc, scale, low, high, validate_args, allow_nan_stats, name)
        super().__init__(distribution, **kwargs)

    def sample(self, *args, **kwargs):
        s = self.distribution.sample(*args, **kwargs)
        low = self.distribution.low
        return tf.maximum(low, s)

    def _tf_moment_4(self, high=None):
        from tensorflow_probability.python.internal.special_math import ndtr
        if high is None:
            high = self.distribution.high
        a,b = self.distribution.low,high
        mu,sigma = self.distribution.loc, self.distribution.scale
        z_b = (b-mu)/sigma
        z_a = (a-mu)/sigma
        
        norm = tfd.Normal(0., 1.) #Standard Normal
        if b==np.inf:
            bterm=0.
        else:
            bterm = (b*b*b + b*b*mu + b*mu*mu + sigma*sigma*(3*b + 5*mu) + mu*mu*mu)*norm.prob(z_b) 
        aterm = (a*a*a+a*a*mu+a*mu*mu+sigma*sigma*(3*a+5*mu)+mu*mu*mu)*norm.prob(z_a)

        num = bterm - aterm
        den = ndtr(z_b) - ndtr(z_a)
        return mu*mu*mu*mu + 6*mu*mu*sigma*sigma + 3*sigma*sigma*sigma*sigma- sigma*num/den

    def _scipy_moment_4(self, high):
        from scipy.stats import truncnorm
        loc,scale = self.distribution.loc, self.distribution.scale
        low = self.distribution.low.numpy()
        if high is None:
            high = self.distribution.high.numpy()
        a, b = (low - loc) / scale, (high - loc) / scale
        mom4 = truncnorm.moment(4, a, b, loc, scale)
        return mom4

    def moment_4(self, high=np.inf, method='scipy'):
        """
        Calculate the fourth moment of this distribution. This is based on the formula here: 
        https://people.smp.uq.edu.au/YoniNazarathy/teaching_projects/studentWork/EricOrjebin_TruncatedNormalMoments.pdf

        Parameters
        ----------
        high : float (optional)
            The high parameter to use for the distribution. By default use inf.
        method : str (optional)
            Either 'scipy' or 'tf'
        """
        if method=='scipy':
            return self._scipy_moment_4(high)
        elif method == 'tf':
            return self._tf_moment_4(high)
        else:
            raise ValueError(f"Unknown method {method} for computing moment_4")

    @classmethod
    def from_loc_and_scale(cls, loc, scale, low=0., high=1e10, scale_shift=1e-7):
        """
        Instantiate a learnable distribution with good default bijectors.

        loc : array
            The initial location of the distribution
        scale : array
            The initial scale parameter of the distribution
        low : float or array (optional)
            The lower limit of the support for the distribution.
        high : float or array (optional)
            The upper limit of the support for the distribution.
        scale_shift : float (optional)
            A small constant added to the scale to increase numerical stability.
        """
        loc   = tfp.util.TransformedVariable(
            loc,
            tfb.Exp(),
        )
        scale = tfp.util.TransformedVariable(
            scale,
            tfb.Chain([
                tfb.Shift(scale_shift),
                tfb.Exp(),
            ]),
        )
        return cls(loc, scale, low, high)

class RiceWoolfson(tfd.Distribution):
    def __init__(self, loc, scale, centric):
        """
        This is a hybrid distribution to parameterize posteriors over structure factors. 
        It uses the Rice distribution to model acentric structure factors and
        the Folded normal or "Woolfson" distribution to model centrics. 

        Parameters
        ----------
        loc : array (float)
            location parameter for the distributions
        scale : array (float)
            scale parameter for the distributions
        centric : array (float;bool)
            Array that same length as loc and scale that is 1./True for centric reflections and 0./False
        """
        self._loc   = tensor_util.convert_nonref_to_tensor(loc, dtype=tf.float32)
        self._scale = tensor_util.convert_nonref_to_tensor(scale, dtype=tf.float32)
        self._centric = np.array(centric, dtype=bool)
        self._woolfson = FoldedNormal(self._loc, self._scale)
        self._rice = Rice(self._loc, self._scale)
        self.eps = np.finfo(np.float32).eps

    def mean(self):
        return tf.where(self._centric, self._woolfson.mean(), self._rice.mean())

    def variance(self):
        return tf.where(self._centric, self._woolfson.variance(), self._rice.variance())

    def stddev(self):
        return tf.where(self._centric, self._woolfson.stddev(), self._rice.stddev())

    def sample(self, sample_shape=(), seed=None, name='sample', **kwargs):
        return tf.where(self._centric, self._woolfson.sample(sample_shape, seed, name, **kwargs)+self.eps, self._rice.sample(sample_shape, seed, name, **kwargs))

    def log_prob(self, x):
        return tf.where(self._centric, self._woolfson.log_prob(x), self._rice.log_prob(x))

    def prob(self, x):
        return tf.where(self._centric, self._woolfson.prob(x), self._rice.prob(x))






#!/usr/bin/env python


def main():
    from . import __version__
    print(f"Careless version {__version__}")
    from careless.parser import parser
    parser = parser.parse_args()
    run_careless(parser)

def run_careless(parser):
    # We defer all inputs to make sure the parser has priority in modifying tf parameters
    import tensorflow as tf
    import numpy as np
    import reciprocalspaceship as rs
    from careless.io.manager import DataManager
    from careless.io.formatter import MonoFormatter,LaueFormatter
    from careless.models.base import BaseModel
    from careless.models.merging.surrogate_posteriors import TruncatedNormal
    from careless.models.merging.variational import VariationalMergingModel
    from careless.models.scaling.image import HybridImageScaler,ImageScaler
    from careless.models.scaling.nn import MLPScaler

    if parser.type == 'poly':
        df = LaueFormatter.from_parser(parser)
    elif parser.type == 'mono':
        df = MonoFormatter.from_parser(parser)
    elif parser.type == 'devices':
        print("###############################################")
        print("# TensorFlow can access the following devices #")
        print("###############################################")
        for dev in tf.config.list_physical_devices():
            print(f" - {dev.device_type}: {dev.name}")
        from sys import exit
        exit()


    inputs,rac = df.format_files(parser.reflection_files)
    dm = DataManager(inputs, rac, parser=parser)

    if parser.test_fraction is not None:
        train,test = dm.split_data_by_refl(parser.test_fraction)
    else:
        train,test = dm.inputs,None

    model = dm.build_model()

    if parser.scale_file is not None:
        model.scaling_model.load_weights(parser.scale_file)
    if parser.freeze_scales:
        model.scaling_model.trainable = False

    if parser.structure_factor_file is not None:
        model.surrogate_posterior.load_weights(parser.structure_factor_file)
    if parser.freeze_structure_factors:
        model.surrogate_posterior.trainable = False

    validation_frequency = parser.validation_frequency
    progress = not parser.disable_progress_bar

    history = model.train_model(
        tuple(map(tf.convert_to_tensor, train)),
        parser.iterations,
        message="Training",
        validation_data=test,
        validation_frequency=validation_frequency,
        progress=progress,
        reduce_retracing=parser.reduce_retracing,
        jit_compile=parser.jit_compile,
    )

    for i,ds in enumerate(dm.get_results(model.surrogate_posterior, inputs=train)):
        filename = parser.output_base + f'_{i}.mtz'
        ds.write_mtz(filename)

    filename = parser.output_base + f'_history.csv'
    history = rs.DataSet(history).to_csv(filename, index_label='step')

    model.surrogate_posterior.save_weights(parser.output_base + '_structure_factor')
    model.scaling_model.save_weights(parser.output_base + '_scale')
    if parser.save_data_manager:
        import pickle
        with open(parser.output_base + "_data_manager.pickle", "wb") as out:
            pickle.dump(dm, out)

    predictions_data = None
    if test is not None:
        for file_id, (ds_train, ds_test) in enumerate(zip(
                dm.get_predictions(model, train, test_value=0),
                dm.get_predictions(model, test, test_value=1),
                )):
            filename = parser.output_base + f'_predictions_{file_id}.mtz'
            rs.concat((
                ds_train,
                ds_test,
            )).write_mtz(filename)
    else:
        for file_id, ds_train in enumerate(dm.get_predictions(model, train, test_value=0)):
            filename = parser.output_base + f'_predictions_{file_id}.mtz'
            ds_train.write_mtz(filename)

    if parser.merge_half_datasets:
        scaling_model = model.scaling_model
        scaling_model.trainable = False
        xval_data = [None] * len(dm.asu_collection)
        for repeat in range(parser.half_dataset_repeats):
            for half_id, half in enumerate(dm.split_data_by_image()):
                model = dm.build_model(scaling_model=scaling_model)
                history = model.train_model(
                    tuple(map(tf.convert_to_tensor, half)), 
                    parser.iterations,
                    message=f"Merging repeat {repeat+1} half {half_id+1}",
                    progress=progress,
                    reduce_retracing=parser.reduce_retracing,
                    jit_compile=parser.jit_compile,
                )

                for file_id,ds in enumerate(dm.get_results(model.surrogate_posterior, inputs=half)):
                    ds['repeat'] = rs.DataSeries(repeat, index=ds.index, dtype='I')
                    ds['half'] = rs.DataSeries(half_id, index=ds.index, dtype='I')
                    if xval_data[file_id] is None:
                        xval_data[file_id] = ds
                    else:
                        xval_data[file_id] = rs.concat((xval_data[file_id], ds))

        for file_id, ds in enumerate(xval_data):
            filename = parser.output_base + f'_xval_{file_id}.mtz'
            ds.write_mtz(filename)

    if parser.embed:
        from IPython import embed
        embed(colors='Linux')


if __name__=="__main__":
    main()




# Version number for careless
def getVersionNumber():
    version = None
    try:
        from setuptools.version import metadata

        version = metadata.version("careless")
    except ImportError:
        from setuptools.version import pkg_resources

        version = pkg_resources.require("careless")[0].version

    return version

__version__ = getVersionNumber()



name = "Scaling Model"
description = """
Options related to the neural network scaling model used for merging. 
"""


args_and_kwargs = (
    (("--scale-file",), {
        "help": "Initialize the scale model weights from the ouput of a previous run. This argument should be a string beginning with the "
                "base filename used in the previous run and ending in _scale.  For instance, if the previous run "
                "was called with `careless mono [...] merge/hewl`, the appropriate file name would be merge/hewl_scale. ",
        "type": str,
        "default": None,
    }),

    (("--freeze-scales",), {
        "help": "Do not optimize the scale model weights.",
        "action": "store_true"
    }),

    (("--mlp-layers",), {
        "help": "The number of dense neural network layers in the scaling model. The default is 20 layers.",
        "type":int,
        "default":20,
    }),

    (("--mlp-width",), {
        "help": "The width of the hidden layers of the neural net. The default is 10.",
        "type": int,
        "default": 10,
    }),

    (("--image-layers",), {
        "help": "Add additional layers with local image-specific parameters.",
        "type":int,
        "default": 0,
    }),


    (("--disable-image-scales",), {
        "help": "Do not learn a local scale param for each image.",
        "action": "store_false",
        "dest" : "use_image_scales",
        "default": True,
    }),

    (("--scale-bijector",), {
        "help": "What function to use to ensure positivity of the standard deviation of scales. "
                "Supported functions are --scale-bijector=exp and the default is --scale-bijector=softplus",
        "type": str,
        "default": "softplus",
        "choices" : ["exp", "softplus"],
    }),
)



import tensorflow as tf
import tf_keras as tfk
from careless.models.base import BaseModel
from careless.models.scaling.base import Scaler
import tensorflow_probability as tfp
import numpy as np


class ImageScaler(Scaler):
    """
    Simple linear image scales. Average value pegged at 1.
    """
    def __init__(self, max_images):
        """
        Parameters
        ----------
        max_images : int
            The maximum number of image variables to be learned
        """
        super().__init__()
        self._scales = tf.Variable(tf.ones(max_images - 1))

    @property
    def scales(self):
        return tf.concat(([1.], self._scales), axis=-1)

    def call(self, inputs):
        """
        Parameters
        ----------
        inputs : list or tf.data.DataSet
            A list of tensor inputs or a DataSet in the standard 
            careless format.

        Returns
        -------
        scales : tf.Tensor(float32)
            A tensor the same shape as image_ids.
        """
        image_ids = self.get_image_id(inputs)
        w = self.scales
        return tf.squeeze(tf.gather(w, image_ids))

class HybridImageScaler(Scaler):
    """
    A scaler that combines an `ImageScaler` with an `MLPScaler`
    """
    def __init__(self, mlp_scaler, image_scaler):
        super().__init__()
        self.mlp_scaler = mlp_scaler
        self.image_scaler = image_scaler

    def call(self, inputs):
        """
        Parameters
        ----------
        """
        q = self.mlp_scaler(inputs)
        a = self.image_scaler(inputs)
        return tfp.distributions.TransformedDistribution(
            q,
            tfp.bijectors.Scale(scale=a),
        )


class ImageLayer(Scaler):
    def __init__(self, units, max_images, activation=None, **kwargs):
        super().__init__(**kwargs)
        self.activation = tfk.activations.get(activation)
        self.units = units
        self.max_images = max_images

    def build(self, input_shape):
        def initializer(shape, dtype=tf.float32, **kwargs):
            return tf.eye(shape[1], shape[2], (shape[0],), dtype=dtype)

        self.w = self.add_weight(
            name='kernel',
            shape=(self.max_images, self.units, input_shape[0][-1]),
            initializer=initializer,
            trainable=True,
        )
        self.b = self.add_weight(
            name='bias', 
            shape=(self.max_images, self.units),
            initializer='zeros',
            trainable=True,
        )

    def call(self, metadata_and_image_id, *args, **kwargs):
        data,image_id = metadata_and_image_id
        image_id = tf.squeeze(image_id)
        w = tf.gather(self.w, image_id, axis=0)
        b = tf.gather(self.b, image_id, axis=0)
        result = self.activation(tf.squeeze(tf.matmul(w, data[...,None]), axis=-1) + b)
        return result

class NeuralImageScaler(Scaler):
    def __init__(self, image_layers, max_images, mlp_layers, mlp_width, leakiness=0.01, epsilon=1e-7, scale_bijector=None, scale_multiplier=None):
        super().__init__()
        layers = []
        if leakiness is None:
            activation = 'ReLU'
        else:
            activation = tfk.layers.LeakyReLU(leakiness)

        for i in range(image_layers):
            layers.append(
                ImageLayer(mlp_width, max_images, activation)
            )

        self.image_layers = layers
        from careless.models.scaling.nn import MetadataScaler
        self.metadata_scaler = MetadataScaler(mlp_layers, mlp_width, leakiness, epsilon=epsilon, scale_bijector=scale_bijector, scale_multiplier=scale_multiplier)

    def call(self, inputs):
        result = self.get_metadata(inputs)
        image_id = self.get_image_id(inputs),

        result = self.metadata_scaler.network(result)

        for layer in self.image_layers:
            result = layer((result, image_id))
        result = self.metadata_scaler.distribution(result)
        return result




import pytest
import reciprocalspaceship as rs
from tempfile import TemporaryDirectory
from careless.careless import run_careless as _run_careless
from os.path import exists
import multiprocessing

niter = 10
# True will use multiprocessing to bypass the tf memory leak issue,
# but it may result in less verbose error messages.
use_mp = False
eager=True


def run_careless(parser):
    """
    Workaround for tensorflow memory leak.
    see: https://github.com/tensorflow/tensorflow/issues/36465
    """
    if eager:
        parser.run_eagerly=True
    if use_mp:
        proc = multiprocessing.Process(target=_run_careless, args=(parser,))
        proc.start()
        proc.join()
    else:
        _run_careless(parser)

def base_test_separate(flags, filenames):
    with TemporaryDirectory() as td:
        out = td + '/out'
        command = flags +  f" --separate-files {' '.join(filenames)} {out}"
        from careless.parser import parser
        parser = parser.parse_args(command.split())
        run_careless(parser)
        for i,in_file in enumerate(filenames):
            out_file = out + f"_{i}.mtz"
            assert exists(out_file)
            in_ds = rs.read_mtz(in_file)
            out_ds = rs.read_mtz(out_file)
            assert in_ds.spacegroup == out_ds.spacegroup
            if parser.dmin is not None:
                assert out_ds.compute_dHKL().dHKL.min() >= parser.dmin
            if parser.anomalous:
                assert 'F(+)' in out_ds

def base_test_together(flags, filenames):
    with TemporaryDirectory() as td:
        out = td + '/out'
        command = flags +  f" {' '.join(filenames)} {out}"
        from careless.parser import parser
        parser = parser.parse_args(command.split())
        run_careless(parser)

        out_file = out + f"_0.mtz"
        assert exists(out_file)
        out_ds = rs.read_mtz(out_file)
        if parser.dmin is not None:
            assert out_ds.compute_dHKL().dHKL.min() >= parser.dmin
        if parser.anomalous:
            assert 'F(+)' in out_ds

@pytest.mark.parametrize("ev11", [True, False])
@pytest.mark.parametrize("dmin", [None, 7.])
@pytest.mark.parametrize("anomalous", [True, False])
@pytest.mark.parametrize("isigi_cutoff", [None, 1.])
@pytest.mark.parametrize("studentt_dof", [None, 12.])
@pytest.mark.parametrize("separate", [True, False])
@pytest.mark.parametrize("mode", ['mono', 'poly'])
@pytest.mark.parametrize("change_sg", [True, False])
def test_twofile(off_file, on_file, on_file_alt_sg, ev11, dmin, anomalous, isigi_cutoff, studentt_dof, separate, mode, change_sg):
    if change_sg:
        on_file = on_file_alt_sg
    flags = f"{mode} --disable-gpu --iterations={niter} dHKL,image_id"
    if ev11:
        flags += f" --refine-uncertainties "
    if dmin is not None:
        flags += f" --dmin={dmin} "
    if anomalous:
        flags += f" --anomalous "
    if isigi_cutoff is not None:
        flags += f" --isigi-cutoff={isigi_cutoff} "
    if studentt_dof is not None:
        flags += f" --studentt-likelihood-dof={studentt_dof} "
    if separate:
        base_test_separate(flags, [off_file, on_file])
    elif change_sg:
        pass #This is an invalid combination. cannot merge different spacegroups into one file.
    else:
        base_test_together(flags, [off_file, on_file])

@pytest.mark.parametrize("mode", ['mono', 'poly'])
def test_double_wilson(off_file, on_file, mode):
    flags = f"{mode} --disable-gpu --iterations={niter} dHKL,image_id"
    flags += " --double-wilson-parents=None,0 "
    r_in_range = " --double-wilson-r=0.0,0.9 "
    r_outside_range = " --double-wilson-r=0.0,1.0 "
    base_test_separate(
        flags + r_in_range, 
        [off_file, on_file]
    )

    with pytest.raises(ValueError):
        base_test_separate(
            flags + r_outside_range, 
            [off_file, on_file]
        )

def test_crystfel(stream_file):
    flags = f"mono --disable-gpu --iterations={niter} --spacegroups=1 dHKL,image_id"
    base_test_together(flags, [stream_file])

    #Careless poly should fail with a clear error message
    with pytest.raises(ValueError):
        flags = f"poly --disable-gpu --iterations={niter} --spacegroups=1 dHKL,image_id"
        base_test_together(flags, [stream_file])

def test_scale_weight_save_and_load(off_file):
    """
    Test saving and loading weights. 
    """
    with TemporaryDirectory() as td:
        out = td + '/out'
        flags = f"mono --disable-gpu --iterations={niter} dHKL,image_id"
        command = flags +  f" {off_file} {out}"
        from careless.parser import parser
        parser = parser.parse_args(command.split())
        run_careless(parser)

        out_file = out + f"_0.mtz"
        assert exists(out_file)

        flags = flags + f" --scale-file={out}_scale"
        out = td + '/out_reloaded'
        command = flags +  f" {off_file} {out}"
        from careless.parser import parser
        parser = parser.parse_args(command.split())
        run_careless(parser)
        out_file = out + f"_0.mtz"
        assert exists(out_file)


def test_structure_factor_save_and_load(off_file):
    """
    Test saving and loading weights. 
    """
    with TemporaryDirectory() as td:
        out = td + '/out'
        flags = f"mono --disable-gpu --iterations={niter} dHKL,image_id"
        command = flags +  f" {off_file} {out}"
        from careless.parser import parser
        parser = parser.parse_args(command.split())
        run_careless(parser)

        out_file = out + f"_0.mtz"
        assert exists(out_file)

        flags = flags + f" --structure-factor-file={out}_structure_factor"
        out = td + '/out_reloaded'
        command = flags +  f" {off_file} {out}"
        from careless.parser import parser
        parser = parser.parse_args(command.split())
        run_careless(parser)
        out_file = out + f"_0.mtz"
        assert exists(out_file)

def test_freeze_structure_factor(off_file):
    """ Test `--freeze-structure-factors` for execution """
    with TemporaryDirectory() as td:
        out = td + '/out'
        flags = f"mono --disable-gpu --iterations={niter} --freeze-structure-factors dHKL,image_id"
        command = flags +  f" {off_file} {out}"
        from careless.parser import parser
        parser = parser.parse_args(command.split())
        run_careless(parser)

        out_file = out + f"_0.mtz"
        assert exists(out_file)

def test_freeze_scales(off_file):
    """ Test `--freeze-scales` for execution """
    with TemporaryDirectory() as td:
        out = td + '/out'
        flags = f"mono --disable-gpu --iterations={niter} --freeze-scales dHKL,image_id"
        command = flags +  f" {off_file} {out}"
        from careless.parser import parser
        parser = parser.parse_args(command.split())
        run_careless(parser)

        out_file = out + f"_0.mtz"
        assert exists(out_file)

@pytest.mark.parametrize('clip_type', ['--clipvalue', '--clipnorm', '--global-clipnorm'])
def test_clipping(off_file, clip_type):
    """ Test gradient clipping settings """
    with TemporaryDirectory() as td:
        out = td + '/out'
        flags = f"mono --disable-gpu --iterations={niter} {clip_type}=1. dHKL,image_id"
        command = flags +  f" {off_file} {out}"
        from careless.parser import parser
        parser = parser.parse_args(command.split())
        run_careless(parser)

        out_file = out + f"_0.mtz"
        assert exists(out_file)


@pytest.mark.parametrize('scale_bijector', ['exp', 'softplus'])
@pytest.mark.parametrize('image_layers', [None, 2])
def test_scale_bijector(off_file, scale_bijector, image_layers):
    """ Test scale bijector settings """
    with TemporaryDirectory() as td:
        out = td + '/out'
        if image_layers is not None:
            flags = f"mono --disable-gpu --image-layers={image_layers} --iterations={niter} --scale-bijector={scale_bijector} dHKL,image_id"
        else:
            flags = f"mono --disable-gpu --iterations={niter} --scale-bijector={scale_bijector} dHKL,image_id"

        command = flags +  f" {off_file} {out}"
        from careless.parser import parser
        parser = parser.parse_args(command.split())
        run_careless(parser)

        out_file = out + f"_0.mtz"
        assert exists(out_file)





import reciprocalspaceship as rs
from careless.models.merging.surrogate_posteriors import RiceWoolfson
from careless.utils.distributions import Stacy
from careless.models.priors.base import Prior
from tensorflow_probability import distributions as tfd
from tensorflow_probability import bijectors as tfb
import tensorflow as tf
import numpy as np


class Centric(tfd.HalfNormal, Prior):
    def __init__(self, epsilon, sigma=1.):
        self.epsilon = tf.convert_to_tensor(epsilon)
        self.sigma = tf.convert_to_tensor(sigma)
        super().__init__(tf.math.sqrt(epsilon * self.sigma))


class Acentric(tfd.Weibull, Prior):
    def __init__(self, epsilon, sigma=1.):
        self.epsilon = tf.convert_to_tensor(epsilon)
        self.sigma = tf.convert_to_tensor(sigma)
        super().__init__(
            2., 
            tf.math.sqrt(self.epsilon * self.sigma),
        )

class WilsonPrior(Prior):
    """Wilson's priors on structure factor amplitudes."""
    def __init__(self, centric, epsilon, sigma=1.):
        """
        Parameters
        ----------
        centric : array
            Floating point or boolean array with value 1/True for centric reflections and 0/False. for acentric.
        epsilon : array
            Floating point array with multiplicity values for each structure factor.
        sigma : float or array
            The Σ value for the wilson distribution. The represents the average intensity stratified by a measure
            like resolution. 
        """
        self.epsilon = np.array(epsilon, dtype=np.float32)
        self.centric = np.array(centric, dtype=bool)
        self.sigma = np.array(sigma, dtype=np.float32)
        self.p_centric = Centric(self.epsilon, self.sigma)
        self.p_acentric = Acentric(self.epsilon, self.sigma)

    def log_prob(self, x):
        """
        Parameters
        ----------
        x : tf.Tensor
            Array of structure factor values with the same shape epsilon and centric.
        """
        return tf.where(self.centric, self.p_centric.log_prob(x), self.p_acentric.log_prob(x))

    def prob(self, x):
        """
        Parameters
        ----------
        x : tf.Tensor
            Array of structure factor values with the same shape epsilon and centric.
        """
        return tf.where(self.centric, self.p_centric.prob(x), self.p_acentric.prob(x))

    def mean(self):
        return tf.where(self.centric, self.p_centric.mean(), self.p_acentric.mean())

    def stddev(self):
        return tf.where(self.centric, self.p_centric.stddev(), self.p_acentric.stddev())

    def sample(self, *args, **kwargs):
        #### BLEERRRGG #####
        return tf.where(
            self.centric, 
            self.p_centric.sample(*args, **kwargs),
            self.p_acentric.sample(*args, **kwargs),
        )

class DoubleWilsonPrior(Prior):
    def __init__(self, asu_collection, parents, r_values, reindexing_ops=None, sigma=1.):
        """
        asu_collection : AsuCollection
        parents : list
            List of integers such that parents[i] = j implies that asu_id==i has parent asu_id==j
        r_values : list or array
            Either a list with the same length as parents or an array that is length == len(asu_collection.lookup_table)
        reindexing_ops : list or tuple
            A list of gemm.Op instances that is the same length as parents.
        sigma : float or array
        """
        self.parents = parents
        self.r_values = r_values
        reflids = []
        loc = []
        scale = []
        root = []
        r = []

        if len(r_values) == len(asu_collection):
            for r_value,asu in zip(r_values, asu_collection):
                r.append(r_value*np.ones(len(asu.lookup_table), dtype='float32'))
            self.r = np.concatenate(r)
        elif len(r_values) == len(asu_collection.lookup_table):
            self.r = r_values
        else:
            raise ValueError("r_values must either one per asu or one per reflection in the collection")

        for child,parent in enumerate(parents):
            child_asu  = asu_collection.reciprocal_asus[child]
            child_size = len(child_asu.lookup_table)
            r_value = r_values[child]

            if parent is None:
                reflids.append(child_asu.lookup_table.id.to_numpy('int32'))
                root.append(np.ones(len(child_asu.lookup_table), dtype='bool'))
            else:
                root.append(np.zeros(len(child_asu.lookup_table), dtype='bool'))
                parent_asu = asu_collection.reciprocal_asus[parent]
                h = child_asu.Hall
                if reindexing_ops is not None:
                    op = reindexing_ops[child]
                    h = rs.utils.apply_to_hkl(h, op)
                h,_ = rs.utils.hkl_to_asu(h, parent_asu.spacegroup)
                pid = parent*np.ones((len(h), 1), dtype='int32')
                reflids.append(asu_collection.to_refl_id(pid, h, allow_missing=True))

        self.centric = asu_collection.centric
        self.multiplicity = asu_collection.multiplicity

        self.sigma = sigma
        self.reflids = np.concatenate(reflids)
        self.absent = tf.convert_to_tensor(self.reflids == -1)
        self.r = np.concatenate(r)
        self.root = np.concatenate(root)
        self.wilson_prior = WilsonPrior(asu_collection.centric, asu_collection.multiplicity, sigma)

    def mean(self):
        return self.wilson_prior.mean()

    def stddev(self):
        return self.wilson_prior.stddev()

    def log_prob(self, z):
        mask = self.reflids >= 0
        sanitized_reflids = tf.where(mask, self.reflids, 0)
        z_parent = tf.where(
            mask[None,:], 
            tf.gather(z, sanitized_reflids, axis=-1),
            0.,
        )

        loc = tf.where(
            self.absent, 
            0.,
            z_parent * self.r
        )
        r2 = tf.square(self.r)
        scale = tf.where(
            self.centric,
            tf.math.sqrt(self.multiplicity * self.sigma * (1. - r2)),
            tf.math.sqrt(0.5*self.multiplicity * self.sigma * (1. - r2)),
        )

        rice_woolfson = RiceWoolfson(loc, scale, self.centric)
        p_wilson = self.wilson_prior.log_prob(z)
        p_dw = rice_woolfson.log_prob(z)
        log_p = tf.where(self.root, p_wilson, p_dw)
        return log_p




name = "Data Filtration"
description = None

args_and_kwargs = (
    # Required args
    (("-c", "--isigi-cutoff"), {
        "help":"Minimum I over Sigma(I) for included reflections. Default is to include all reflections", 
        "type":float, 
        "default":None,
    }),

    (("-d", "--dmin"), {
        "help":f"Maximum resolution in Ångstroms. If this is not supplied," 
                "reflections will be merged out to the highest resolution reflection present in the input.", 
        "type":float, 
        "default":None,
    }),
)



from tqdm.keras import TqdmCallback
from tqdm.auto import tqdm

class auto_tqdm(tqdm):
    def __init__(self, *args, numeric_format='{0:.3e}', **kwargs):
        super().__init__(*args, **kwargs)
        self.numeric_format=numeric_format

    def format_num(self, n):
        return self.numeric_format.format(n)

class ProgressBar(TqdmCallback):
    """ Keras Callback based on tqdm.keras.TqdmCallback but with fixed width metrics. """
    def __init__(self, *args, tqdm_class=auto_tqdm, verbose=0, **kwargs):
        super().__init__(*args, tqdm_class=auto_tqdm, verbose=verbose, **kwargs)




name = "Crossvalidation"
description = """
Careless supports two sorts of crossvalidation. The first is used to assess the validity of a particular merging
strategy (see Model Selection). The second is meant to assess the quality of a data set (see Data Consistency). 
By default, careless will run neither. However, You can use any combination of these strategies using the provided flags. \n

  Model Selection\n
  ---------------\n
  Use the `--test-fraction` option to assess model fit and diagnose overfitting. This method will reserve a fraction of the 
  data into a test set, and output the predictions for each reflection observation in the training and test sets at the end
  of merging. \n

  Data Consistency\n
  ----------------\n
  Careless is able to provide half dataset merging results for assessing data resolution and quality. Setting the 
  `--merge-half-datasets` flag will first train the model on the full training set of data. Afterwards, the data will
  be split into two halves by randomly partitioning the images. With the model weights frozen, structure factors 
  are estimated by optimizing the loss function for each half dataset. The half dataset merging results will be written
  to a file which can be analyzed to estimate conventional statistics such as CChalf. One file will be written for each
  merged `mtz`. These files have the suffix *_xval_#.mtz 

"""


args_and_kwargs = (
    (
        ("--test-fraction",),
        {
            "help": "Output model predictions for a held-out fraction of data. This should be used for model selection purposes. "
            "By default, no data will be held out during training. ",
            "type": float,
            "default": None,
        },
    ),
    (
        ("--merge-half-datasets",),
        {
            "help": "After training, split the data in half randomly by image and merge each half using the scaling model learned on the training fraction. "
            "The output of the halves will be written to a file which can be used to estimate traditional CChalf type measures. The full data set will"
            " always be used to generate half data sets irrespective of the test fraction. One file is written for each merged mtz. The files have the "
            "*_xval_#.mtz suffix. ",
            "action": "store_true",
            "default": False,
        },
    ),
    (
        ("--half-dataset-repeats",),
        {
            "help": "Number of times to Repeat the half dataset crossvalidation. By default this is one.",
            "type": int,
            "default": 1,
        },
    ),
    (
        ("--validation-frequency",),
        {
            "help": "During training, set how frequently to evaluate the model on the test set. This is an integer >= 1 which defaults to 10 for once every 10 steps.",
            "type": int,
            "default": 10,
        },
    ),
)



import tensorflow as tf
from tensorflow_probability.python.internal import special_math
from tensorflow_probability import distributions as tfd
import tensorflow_probability as tfp
from tensorflow_probability import bijectors as tfb
from tensorflow_probability.python.internal import tensor_util
import numpy as np


class Amoroso(tfd.TransformedDistribution):
    def __init__(self,
		   a,
		   theta,
		   alpha,
                   beta,
		   validate_args=False,
		   allow_nan_stats=True,
		   name='Amoroso'):

        parameters = dict(locals())
        with tf.name_scope(name) as name:
            self._a = tensor_util.convert_nonref_to_tensor(a)
            self._theta = tensor_util.convert_nonref_to_tensor(theta)
            self._alpha = tensor_util.convert_nonref_to_tensor(alpha)
            self._beta = tensor_util.convert_nonref_to_tensor(beta)
            gamma = tfd.Gamma(alpha, 1.)

            chain = tfb.Invert(tfb.Chain([
                tfb.Exp(),
                tfb.Scale(beta),
                tfb.Shift(-tf.math.log(theta)),
                tfb.Log(),
                tfb.Shift(-a),
            ]))

            super().__init__(
                    distribution=gamma, 
                    bijector=chain, 
                    validate_args=validate_args,
                    parameters=parameters,
                    name=name)

    @property
    def a(self):
        return self._a

    @property
    def theta(self):
        return self._theta

    @property
    def alpha(self):
        return self._alpha

    @property
    def beta(self):
        return self._beta

    def custom_log_prob(self, X):
        a = self.a
        theta = self.theta
        alpha = self.alpha
        beta  = self.beta

        inflation = 1e10
        x = tf.math.log(tf.abs(beta)) - tf.math.log(tf.abs(theta)) - tf.math.lgamma(a) 
        y = (alpha*beta - 1.)*(tf.math.log(inflation*X - inflation*a) - tf.math.log(theta) - tf.math.log(inflation)) 
        z = - tf.math.pow(X - a, beta) * tf.math.exp(- beta * tf.math.log(theta))
        print(x,y,z)

        return x + y + z

    def mean(self):
        """
        The mean of of the Amoroso distribution exists for `alpha + 1/beta >= 0`.
        It can be computed analytically by

        ```
        mean = a + theta * gamma(alpha + 1/beta) / gamma(alpha)
        ```
        """
        a,theta,alpha,beta = self.a,self.theta,self.alpha,self.beta
        return a + tf.math.exp(tf.math.log(theta)+tf.math.lgamma(alpha + tf.math.reciprocal(beta)) - tf.math.lgamma(alpha))

    def variance(self):
        """
        The variance of of the Amoroso distribution exists for `alpha + 2/beta >= 0`.
        It can be computed analytically by

        ```
        variance = theta**2 * (gamma(alpha + 2/beta) / gamma(alpha) - gamma(alpha + 1/beta)**2 / gamma(alpha)**2 )
        ```
        """
        theta,alpha,beta = self.theta,self.alpha,self.beta
        return theta**2. * (tf.math.exp(tf.math.lgamma(alpha + 2./beta) - tf.math.lgamma(alpha)) - tf.math.exp(2.*tf.math.lgamma(alpha + 1/beta) - 2.*tf.math.lgamma(alpha)))

    def stddev(self):
        """
        The variance of of the Amoroso distribution exists for `alpha + 2/beta >= 0`.
        It can be computed analytically by

        ```
        variance = theta**2 * (gamma(alpha + 2/beta) / gamma(alpha) - gamma(alpha + 1/beta)**2 / gamma(alpha)**2 )
        ```

        The standard deviation is computed by
        ```amoroso.stddev() = tf.sqrt(amoroso.variance())```
        """
        return tf.math.sqrt(self.variance())


class Stacy(Amoroso):
    def __init__(self,
		 theta,
		 alpha,
                 beta,
		 validate_args=False,
		 allow_nan_stats=True,
		 name='Stacy'):

        parameters = dict(locals())
        with tf.name_scope(name) as name:
            super().__init__(
                0.,
                theta,
                alpha,
                beta,
                validate_args,
                allow_nan_stats,
                name,
            )

    @classmethod
    def wilson_prior(cls, centric, epsilon, Sigma=1.):
        """
        Construct a wilson prior based ont he Stacy distribution.
        Centric Wilson distributions are HalfNormals with scale = sqrt(epsilon * Sigma)
        P(F|Sigma,epsilon)  = (2*pi*Sigma*epsilon)**(-0.5) * exp(-F**2 / 2 / Sigma / epsilon)
                            = Stacy(F| sqrt(2*Sigma*epsilon), 0.5, 2)

        Acentric Wilson distribution 
        P(F|Sigma,epsilon) = (2/Sigma/epsilon) * F * exp(-(F**2/Sigma/epsilon))
                           = Stacy(F | sqrt(Sigma*epsilon), 1., 2.)

        Parameters
        ----------
        centric : array (bool)
            boolean array with True for centric reflections
        epsilon : array (float)
            float array of structure factor multiplicities
        """
        centric = tf.cast(centric, dtype=tf.float32)
        epsilon = tf.convert_to_tensor(epsilon, dtype=tf.float32)
        #centric = np.array(centric, dtype=np.float32) #<-- coerce same same
        theta = centric*np.sqrt(2. * epsilon * Sigma) + (1.-centric)*np.sqrt(Sigma*epsilon)
        alpha = centric*0.5 + (1.-centric)
        beta = centric*2. + (1. - centric)*2.
        return cls(theta, alpha, beta)

    @staticmethod
    def _stacy_params(dist):
        if isinstance(dist, Stacy):
            params = (dist.theta, dist.alpha, dist.beta)
        elif isinstance(dist, tfd.Weibull):
            # Weibul(x; k, lambda) = Stacy(x; lambda, 1, k)
            k = dist.concentration
            lam = dist.scale
            params = (lam, 1., k)
        elif isinstance(dist, tfd.HalfNormal):
            #HalfNormal(x; scale) = Stacy(x; sqrt(2)*scale, 0.5, 2)
            scale = dist.scale
            params = (np.sqrt(2.) * scale, 0.5, 2.)
        else:
            raise TypeError(f"Equivalent Stacy parameters cannot be determined for distribution, {dist}. " 
                             "Only tfd.Weibull, tfd.HalfNormal, or Stacy can be converted to Stacy parameterisation")
        return params

    @staticmethod
    def _bauckhage_params(dist):
        theta, alpha, beta = Stacy._stacy_params(dist)
        bauckhage_params = (theta, alpha*beta, beta)
        return bauckhage_params 

    def kl_divergence(self, other):
        """
        The Stacy distribution has an analytical KL div. 
        However, it isn't documented in the same parameterization as the Crooks Amoroso tome. 
        To avoid confusion, I will first translate the Stacy distributions from 
        Crooks's parameterization to the one in the KL div paper. 

        ```
        Stacy(x; a,d,p) = x**(d-1) * exp(-(x/a)**p) / Z
        ```
        where
        ```
        a = theta
        d = alpha * beta
        p = beta
        ```
        Then the KL div is
        ```
        log(p1*a2**d2*gamma(d2/p2)) - log(p2*a1**d1*gamma(d1/p1)) + 
        (digamma(d1/p1)/p1 + log(a1)) * (d1 - d2) + 
        gamma((d1 + p2)/p1) * (a1/a2)**p2 / gamma(d1/p1) - d1/p1
        ```

        See Bauckhage 2014 for derivation. 
        https://arxiv.org/pdf/1401.6853.pdf

        Parameters
        ----------
        other : Stacy or tfd.Weibull or tfd.HalfNormal
        """
        a1,d1,p1 = self._bauckhage_params(self)
        a2,d2,p2 = self._bauckhage_params(other)

        #The following numerics are easier to read if you alias this
        ln = tf.math.log

        kl = ln(p1) + d2*ln(a2) + tf.math.lgamma(d2/p2) - ln(p2) - d1*ln(a1) - tf.math.lgamma(d1/p1) + \
             (tf.math.digamma(d1/p1)/p1 + ln(a1))*(d1 - d2) +  \
             tf.math.exp(tf.math.lgamma((d1 + p2)/p1) - tf.math.lgamma(d1/p1) + p2*(ln(a1) - ln(a2))) \
             - d1/p1

        return kl


class Rice(tfd.Distribution):
    def __init__(self,
		   nu,
		   sigma,
		   validate_args=False,
		   allow_nan_stats=True,
		   name='Rice'):

        parameters = dict(locals())
        #Value of nu/sigma for which, above with the pdf and moments will be swapped with a normal distribution
        self._normal_crossover = 40. 
        with tf.name_scope(name) as name:
            self._nu = tensor_util.convert_nonref_to_tensor(nu)
            self._sigma = tensor_util.convert_nonref_to_tensor(sigma)
            self._base_normal = tfd.Normal(0., self._sigma)

    @property
    def nu(self):
        return self._nu

    @property
    def sigma(self):
        return self._sigma

    def sample(self, sample_shape=(), seed=None, name='sample', **kwargs):
        s1 = self._base_normal.sample(sample_shape=sample_shape, seed=seed, name=name, **kwargs)
        s2 = self._base_normal.sample(sample_shape=sample_shape, seed=seed, name=name, **kwargs)
        return tf.math.sqrt(
                    s1**2. +
                   (s2 + self.nu)**2.
               )

    def _log_bessel_i0(self, x):
        return tf.math.log(tf.math.bessel_i0e(x)) + tf.math.abs(x)

    def _log_bessel_i1(self, x):
        return tf.math.log(tf.math.bessel_i1e(x)) + tf.math.abs(x)

    def _bessel_i0(self, x):
        return tf.math.exp(self._log_bessel_i0(x))

    def _bessel_i1(self, x):
        return tf.math.exp(self._log_bessel_i1(x))

    def _laguerre_half(self, x):
        return (1. - x) * tf.math.exp(x / 2. + self._log_bessel_i0(-0.5 * x)) - x * tf.math.exp(x / 2.  + self._log_bessel_i1(-0.5 * x) )

    def prob(self, X):
        return tf.math.exp(self.log_prob(X))

    def log_prob(self, X):
        sigma = self.sigma
        nu = self.nu
        log_p = tf.math.log(X) - 2.*tf.math.log(sigma) - (X**2. + nu**2.)/(2*sigma**2.) + \
                    self._log_bessel_i0(X * nu/sigma**2.)
        return log_p

    def mean(self):
        sigma = self.sigma
        nu = self.nu
        mean = sigma * tf.math.sqrt(np.pi / 2.) * self._laguerre_half(-0.5*(nu/sigma)**2)
        return tf.where(nu/sigma > self._normal_crossover,  nu, mean)

    def variance(self):
        sigma = self.sigma
        nu = self.nu
        variance = 2*sigma**2. + nu**2. - 0.5*np.pi * sigma**2. * self._laguerre_half(-0.5*(nu/sigma)**2)**2.
        return tf.where(nu/sigma > self._normal_crossover,  sigma**2., variance)

    def stddev(self):
        return tf.math.sqrt(self.variance())

class FoldedNormal(tfd.TransformedDistribution):
    def __init__(self,
		   loc,
		   scale,
		   validate_args=False,
		   allow_nan_stats=True,
		   name='FoldedNormal'): 
        parameters = dict(locals())
        with tf.name_scope(name) as name:
            self._loc   = tensor_util.convert_nonref_to_tensor(loc)
            self._scale = tensor_util.convert_nonref_to_tensor(scale)
            normal = tfd.Normal(self._loc, self._scale)
            bijector = tfb.AbsoluteValue()

            super().__init__(
                    distribution=normal, 
                    bijector=bijector, 
                    validate_args=validate_args,
                    parameters=parameters,
                    name=name)

    @property
    def loc(self):
        return self._loc

    @property
    def scale(self):
        return self._scale

    def prob(self,  value, name='prob', **kwargs):
        p = super().prob(value, name, **kwargs)
        return tf.where(value < 0, tf.zeros_like(p), p)

    def log_prob(self,  value, name='prob', **kwargs):
        p = super().log_prob(value, name, **kwargs)
        return tf.where(value < 0, np.nan, p)

    def mean(self, name='mean', **kwargs):
        u = self.loc
        s = self.scale
        return s * np.sqrt(2/np.pi) * tf.math.exp(-0.5 * (u/s)**2.) + u * (1. - 2. * special_math.ndtr(-u/s))

    def variance(self, name='variance', **kwargs):
        u = self.loc
        s = self.scale
        return u**2. + s**2. - self.mean()**2.

    def stddev(self, name='stddev', **kwargs):
        return tf.math.sqrt(self.variance())

if __name__=="__main__":
    from careless.models.priors.wilson import Centric,Acentric
    F = np.linspace(1e-10, 10, 1000)
    #Test centrics
    epsilon = np.random.choice([1., 2., 3., 4., 6.], 100).astype(np.float32)
    Sigma = np.random.random(100).astype(np.float32)
    centric = np.ones(100) == 1.
    test = Stacy.wilson_prior(centric, epsilon, Sigma)
    ref = Centric(epsilon, Sigma)
    assert np.all(np.isclose(test.log_prob(F[:,None]), ref.log_prob(F[:,None]), rtol=1e-3))
    assert np.all(np.isclose(test.prob(F[:,None]), ref.prob(F[:,None]), rtol=1e-3))

    #Test acentrics
    centric = np.ones(100) == 0.
    test = Stacy.wilson_prior(centric, epsilon, Sigma)
    ref = Acentric(epsilon, Sigma)
    assert np.all(np.isclose(test.prob(F[:,None]), ref.prob(F[:,None]), rtol=1e-3))

    #Test higher dimensions
    centric = np.random.choice([True, False], [10, 20, 30])
    epsilon = np.random.choice([1., 2., 3., 4., 6.], [10, 20, 30]).astype(np.float32)
    q = Stacy.wilson_prior(centric, epsilon)
    q.log_prob(q.sample())




"""
Convert XDS hkl files to mtzs for careless.
"""

import argparse
import numpy as np
import reciprocalspaceship as rs
import gemmi


class ArgumentParser(argparse.ArgumentParser):
    def __init__(self):
        super().__init__(
            formatter_class=argparse.RawTextHelpFormatter, 
            description=__doc__
        )

        self.add_argument(
            "hkl", 
            help="Unmerged HKL file from XDS.",
        )

        self.add_argument(
            "mtz_out", 
            help="Output mtz file name.",
        )

        self.add_argument(
            "-t", 
            "--file-type", 
            default=None,
            type=str,
            help="Override the type of HKL file. This argument can be one of 'ascii' or 'integrate'. "
                 "The default is None which will try to use the header to infer the type.",
        )

        self.add_argument(
            "-s", 
            "--spacegroup", 
            default=None,
            type=str,
            help="Override the space group. This can be a space group number like 19 or a string like 'P 21 21 21'. ",
        )

        self.add_argument(
            "-c", 
            "--cell", 
            default=None,
            nargs=6,
            metavar=("a", "b", "c", "alpha", "beta", "gamma"),
            type=float,
            help="Override the unit cell. Supply six numbers (e.g. --cell 10 20 30 90 90 90).",
        )


def get_unit_cell(file_name):
    for line in open(file_name):
        if line.startswith("!UNIT_CELL_CONSTANTS="):
            cell = [float(i) for i in line.split()[1:]]
            return gemmi.UnitCell(*cell)

def get_space_group(file_name):
    for line in open(file_name):
        if line.startswith("!SPACE_GROUP_NUMBER="):
            sg = line.split()[1]
            return gemmi.SpaceGroup(sg)

def _read_hkl(file_name, cell=None, spacegroup=None, names=None):

    if cell is None:
        cell = get_unit_cell(file_name)
    if spacegroup is None:
        spacegroup = get_space_group(file_name)

    ds = rs.read_csv(
        file_name,
        delim_whitespace=True,
        comment='!',
        names=names,
        cell=cell,
        spacegroup=spacegroup,
        merged=False,
    )
    return ds

def read_integrate_hkl(file_name, cell=None, spacegroup=None):
    cols = [
        "H", "K", "L",
        "IOBS","SIGMA",
        "XCAL", "YCAL", "ZCAL",
        "RLP", "PEAK", "CORR", "MAXC", 
        "XOBS", "YOBS", "ZOBS", 
        "ALF0","BET0","ALF1","BET1","PSI","ISEG"
    ]
    ds = _read_hkl(file_name, cell, spacegroup, cols)
    ds['BATCH'] = ds.ZOBS.round().astype("B")
    return ds

def get_column_names(file_name):
    cols = None
    for line in open(file_name):
        if line.startswith("!NUMBER_OF_ITEMS_IN_EACH_DATA_RECORD="):
            num_cols = int(line.split('=')[1])
            cols = [str(i) for i in range(num_cols)]
        if line.startswith("!ITEM"):
            name = line.split('_')[1].split("=")[0]
            index = int(line.split("=")[1])
            cols[index - 1] = name
        if line.startswith("!END_OF_HEADER"):
            break
    return cols

def read_ascii_hkl(file_name, cell=None, spacegroup=None, zkey='ZD'):
    cols = get_column_names(file_name)
    ds = _read_hkl(file_name, cell, spacegroup, cols)
    if zkey in ds:
        ds['BATCH'] = ds[zkey].round().astype("B")
    return ds

def get_format_field(file_name):
    for line in open(file_name):
        if line.startswith("!FORMAT="):
            return line.split()[0].split('=')[1]
        if line.startswith('!END_OF_HEADER'):
            break
    return None

def infer_file_type(file_name):
    format_name = get_format_field(file_name)
    if format_name == 'XDS_ASCII':
        return 'ascii'
    elif format_name is None:
        return 'integrate'
    else:
        raise ValueError(f"Could not determine filetype for file_name: {file_name}")

def read_hkl(file_name, cell=None, spacegroup=None, file_type=None):
    if file_type is None:
        file_type = infer_file_type(file_name)

    if file_type == 'integrate':
        return read_integrate_hkl(file_name, cell, spacegroup)
    elif file_type == 'ascii':
        return read_ascii_hkl(file_name, cell, spacegroup)
    else:
        raise ValueError(f"file_type, {file_type} not one of 'integrate', 'ascii'.")

def run(parser):
    ds = read_hkl(
        parser.hkl, 
        parser.cell, 
        parser.spacegroup, 
        parser.file_type
    )
    ds.set_index(['H', 'K', 'L']).write_mtz(parser.mtz_out)

def main():
    parser = ArgumentParser().parse_args()
    run(parser)
